from __future__ import annotations
import asyncio
import datetime
import functools
import traceback

from typing import Callable, Coroutine, TypeVar

import discord
from discord import ui

from lgrez import commons, config
from lgrez.blocs import tools


class _DiscordJourneyView(ui.View):
    def __init__(self, journey: DiscordJourney, timeout: float | None = 180):
        self.journey = journey
        super().__init__(timeout=timeout)

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        return await self.journey._interaction_check(interaction)


class _DiscordJourneyModal(ui.Modal):
    def __init__(self, journey: DiscordJourney, title: str, timeout: float | None = 180):
        self.journey = journey
        super().__init__(title=title, timeout=timeout)

    async def on_submit(self, interaction: discord.Interaction) -> None:
        self.journey.interaction = interaction
        self.stop()

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        return await self.journey._interaction_check(interaction)


def _create_button(input: ui.Button | str | discord.Emoji) -> ui.Button:
    if isinstance(input, ui.Button):
        if input.label:
            input.label = input.label[:80]
        return input
    if isinstance(input, discord.Emoji):
        return ui.Button(emoji=input)
    if len(input) <= 2 and not input.isascii():
        return ui.Button(emoji=input)
    else:
        return ui.Button(label=str(input)[:80])


def _create_select_option(input: discord.SelectOption | str) -> discord.SelectOption:
    if isinstance(input, discord.SelectOption):
        return input
    return discord.SelectOption(label=input)


_Key = TypeVar("_Key")


class DiscordJourney:
    """Home-made protocol to make use of interactions easy when writing application commands.

    Args:
        interaction: The interaction that triggered the command.
        ephemeral: If ``True``, all messages are not public and shown only to the interaction triggerer.
        author_only: The only member allowed to interact ; if not passed, everyone can interact.
        timeout: The delay after which the interaction expires, in seconds.
    """

    RUNNING_INTERACTION: discord.Interaction = None
    TIMEOUT_DELAY: float = 2.0

    _catch_next_command_channel: discord.TextChannel | None = None
    _catch_next_command_member: discord.Member | None = None
    _catch_next_command_coro: Coroutine
    _tasks = set()

    def __init__(
        self,
        interaction: discord.Interaction,
        *,
        ephemeral: bool = False,
        author_only: discord.Member | None = None,
        timeout: float | None = None,
        command_author: discord.Member | None = None,
    ) -> None:
        self._thinking = True

        self.interaction = interaction
        self.ephemeral = ephemeral
        self.author_only = author_only
        self.timeout = timeout

        self.member = interaction.user
        if not isinstance(self.member, discord.Member):
            raise ValueError(f"Interaction user is not a guild member (type {type(self.member)})")
        self.command_author = command_author or self.member

        self.channel: discord.TextChannel = interaction.channel
        self.created_at: datetime.datetime = interaction.created_at

        self._stopped: bool = False

    @property
    def interaction(self) -> discord.Interaction:
        """The active interaction of this journey."""
        return self._interaction

    @interaction.setter
    def interaction(self, new_interaction: discord.Interaction) -> None:
        if not isinstance(new_interaction, discord.Interaction):
            raise TypeError(f"DiscordJourney.interaction must be a discord.Interaction! (got {type(new_interaction)})")
        self._interaction = new_interaction
        if self._thinking:
            asyncio.get_running_loop().call_later(self.TIMEOUT_DELAY, self._timeout_interaction, self.interaction)

    def _timeout_interaction(self, interaction: discord.Interaction) -> None:
        async def _timeout():
            if not interaction.response.is_done():
                await interaction.response.defer(thinking=True, ephemeral=self.ephemeral)

        task = asyncio.create_task(_timeout())
        self._tasks.add(task)  # Cf. https://docs.python.org/fr/3/library/asyncio-task.html#asyncio.create_task
        task.add_done_callback(self._tasks.discard)

    def last_interaction(self):
        self._thinking = False

    async def _send_message(
        self,
        content: str = "",
        view: ui.View | None = None,
        code: bool = False,
        prefix: str = "",
        ephemeral: bool = False,
        **kwargs,
    ) -> list[discord.Message]:
        return await tools.send_blocs(
            self.interaction,
            content,
            code=code,
            prefix=prefix,
            view=view,
            ephemeral=ephemeral or self.ephemeral,
            author=self.command_author,
            **kwargs,
        )

    async def send(self, content: str = "", code: bool = False, prefix: str = "", **kwargs) -> list[discord.Message]:
        """Send a message in this journey.

        Args:
            content: The textual contents to send.
            code: Whether to send the messages as code blocks.
            prefix: Is ``code`` is ``True``, optional text to send before the code blocks.
            kwargs: Passed to ``tools.send_blocs``.

        Returns:
            The messages posted.
        """
        return await self._send_message(content, code=code, prefix=prefix, **kwargs)

    def _check_state(self) -> None:
        if self._stopped:
            raise RuntimeError("Journey is stopped")

    async def _interaction_check(self, interaction: discord.Interaction) -> bool:
        if self.author_only and self.author_only != interaction.user:
            return False
        return True

    async def __aenter__(self):
        DiscordJourney.RUNNING_INTERACTION = None
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if not self.interaction.response.is_done():
            if exc_type:
                DiscordJourney.RUNNING_INTERACTION = self.interaction
            else:
                try:
                    await self.interaction.response.defer()
                except discord.HTTPException:
                    pass
        self._stopped = True

    @staticmethod
    def _clicked_button_cleanup(button: ui.Button, view: ui.View) -> None:
        button.disabled = True

    @staticmethod
    def _not_clicked_button_cleanup(button: ui.Button, view: ui.View) -> None:
        view.remove_item(button)

    async def buttons(
        self,
        content: str,
        buttons: dict[_Key, ui.Button | str | discord.Emoji],
        clicked_button_cleanup: Callable[[ui.Button, ui.View], None] = _clicked_button_cleanup,
        not_clicked_button_cleanup: Callable[[ui.Button, ui.View], None] = _not_clicked_button_cleanup,
        **kwargs,
    ) -> _Key:
        """Send a message with buttons and wait for a click on one of them.

        Args:
            content: The message to send.
            buttons: The buttons to show, indexed by any hashable key, described either with a complete
                ``discord.ui.Button``, a simple label or an emoji.
            clicked_button_cleanup: What to do to the clicked button. Marks it disabled by default.
            not_clicked_button_cleanup: What to do to other buttons. Delete them by default.
            kwargs: Passed to ``journey.send``.

        Returns:
            The key index of the clicked button.
        """
        self._check_state()
        view = _DiscordJourneyView(self, timeout=self.timeout)
        clicked_button = None

        def _get_button_callback(button: ui.Button) -> Callable[[discord.Interaction], Coroutine]:
            async def _callback(interaction: discord.Interaction):
                nonlocal clicked_button
                self.interaction = interaction
                clicked_button = button
                view.stop()

            return _callback

        for key, button in buttons.items():
            button = _create_button(button)
            button.callback = _get_button_callback(button)
            view.add_item(button)
            buttons[key] = button

        *_, message = await self._send_message(content, view, **kwargs)  # Show buttons to user
        try:
            timeout = await view.wait()  # Wait for user click
        except asyncio.CancelledError:
            pass  # Clean view when cancelled
        else:
            if timeout:
                raise RuntimeError("Journey timeout")
        finally:
            # Clean view
            clicked_key = None
            for key, button in buttons.items():
                if button == clicked_button:
                    clicked_key = key
                    clicked_button_cleanup(button, view)
                else:
                    not_clicked_button_cleanup(button, view)
            await message.edit(view=view)

        return clicked_key

    async def yes_no(self, content: str, yes_emoji="✅", no_emoji="❌", **kwargs) -> bool:
        """Send a message with positive and negative buttons and wait for a click on one of them.

        Args:
            content: The message to send.
            yes_emoji: The emoji of the positive button.
            no_emoji: The emoji of the negative button.
            kwargs: Passed to ``journey.buttons``.

        Returns:
            The key index of the clicked button.
        """
        buttons = {
            True: ui.Button(emoji=yes_emoji, style=discord.ButtonStyle.success),
            False: ui.Button(emoji=no_emoji, style=discord.ButtonStyle.secondary),
        }
        return await self.buttons(content, buttons, **kwargs)

    async def ok_cancel(self, content: str, ok_emoji="✅", cancel_emoji="❌", **kwargs) -> None:
        """Send a message with positive and negative buttons, aborts command execution if the latter is clicked.

        Args:
            content: The message to send.
            ok_emoji: The emoji of the positive button.
            cancel_emoji: The emoji of the negative button.
            kwargs: Passed to ``journey.buttons``.

        Raises:
            commons.CommandAbortedError: If the cancel button is clicked.
        """
        ok_button = ui.Button(emoji=ok_emoji, style=discord.ButtonStyle.success)
        cancel_button = ui.Button(emoji=cancel_emoji, style=discord.ButtonStyle.secondary)

        def clicked_button_cleanup(button: ui.Button, view: ui.View) -> None:
            self._clicked_button_cleanup(button, view)
            if button == cancel_button:
                button.label = "Mission aborted."

        if not await self.buttons(
            content, {True: ok_button, False: cancel_button}, clicked_button_cleanup=clicked_button_cleanup, **kwargs
        ):
            await self.interaction.response.defer()
            raise commons.CommandAbortedError()

    async def select(
        self,
        content: str,
        options: dict[_Key, discord.SelectOption | str],
        placeholder: str | None = None,
        validate: str | None = None,
    ) -> _Key:
        """Send a message with a select menu and wait for one choice to be picked.

        Args:
            content: The message to send.
            options: The options of the select menu, indexed by any hashable key, described either with a
                complete ``discord.ui.SelectOption`` or a simple label.
            placeholder: Optional placeholder text, the first text is selected if not passed.
            validate: If passed, add a button next to the select menu with this label, to click on
                to validate the choice.

        Returns:
            The key index of the clicked button.
        """
        self._check_state()
        view = _DiscordJourneyView(self, timeout=self.timeout)

        options = {key: _create_select_option(option) for key, option in options.items()}
        select = ui.Select(options=options.values(), placeholder=placeholder)
        view.add_item(select)

        async def _callback(interaction: discord.Interaction):
            self.interaction = interaction
            view.stop()

        if validate:
            button = ui.Button(label=validate)
            view.add_item(button)
            button.callback = _callback

            async def _noop_callback(interaction: discord.Interaction):
                await interaction.response.defer()

            select.callback = _noop_callback
        else:
            select.callback = _callback

        *_, message = await self._send_message(content, view)  # Show select to user
        try:
            timeout = await view.wait()  # Wait for user click
        except asyncio.CancelledError:
            pass  # Clean view and exit
        else:
            if timeout:
                raise RuntimeError("Journey timeout")

        # Clean view
        if select.values:
            [value] = select.values
            key, option = next((key, option) for key, option in options.items() if option.value == value)
            option.default = True
            select.options = [option]
            select.disabled = True
        else:
            view.remove_item(select)
        if validate:
            view.remove_item(button)
        await message.edit(view=view)

        return key

    async def modal(self, title: str, *items: discord.ui.TextInput | str) -> list[str]:
        """Show a modal and wait for it's completion.

        Args:
            title: The modal title.
            items: The fields to include in the modal, described either with a complete
                ``discord.ui.TextInput`` or a simple label.

        Returns:
            The list of values filled in the fields, in the same order.
        """
        self._check_state()
        modal = _DiscordJourneyModal(self, title=title, timeout=self.timeout)

        inputs = [ui.TextInput(label=item) if isinstance(item, str) else item for item in items]
        for input in inputs:
            modal.add_item(input)

        if self.interaction.response.is_done():
            raise RuntimeError("No active interaction, cannot send modal")

        await self.interaction.response.send_modal(modal)  # Show modal to user
        timeout = await modal.wait()  # Wait for modal completion
        if timeout:
            raise RuntimeError("Journey timeout")
        self._check_state()

        return [input.value for input in inputs]

    async def _catch_next_command(self) -> tuple[discord.Interaction, Callable[[DiscordJourney], Coroutine]]:
        class _Awaitable:
            # Oh, well... let's say it just work?
            # If you really want to dig, begin with https://stackoverflow.com/a/60118660
            def __await__(self):
                while True:
                    result = yield
                    if result:
                        yield
                        return result

        async def _catch_next_command():
            return await _Awaitable()

        self._catch_next_command_channel = self.channel
        self._catch_next_command_member = self.member
        self._catch_next_command_coro = _catch_next_command()

        try:
            return await self._catch_next_command_coro
        finally:  # Stop catching
            self._catch_next_command_channel = None
            self._catch_next_command_member = None
            self._catch_next_command_coro = None

    async def catch_next_command(
        self, next_command_message: str
    ) -> tuple[discord.Interaction, Callable[[DiscordJourney], Coroutine]]:
        """Catch the next command in the channel (or abort it with a cancel button).

        Args:
            next_command_message: The message describing which command to send next.

        Raises:
            commons.CommandAbortedError: If the cancel button is clicked, or if an exception occurs.

        Returns:
            The interaction that triggered the catched command, and a coroutine function to run it.
        """
        catch_task = asyncio.Task(self._catch_next_command())
        cancel_task = asyncio.Task(
            self.buttons(
                f"{next_command_message} :arrow_down:",
                {"cancel": discord.ui.Button(label="Annuler", emoji="❌")},
                ephemeral=True,
            )
        )
        done, _ = await asyncio.wait([catch_task, cancel_task], return_when=asyncio.FIRST_COMPLETED)

        match list(done):
            case [task] if task == catch_task:
                cancel_task.cancel()  # Remove cancel button
                return catch_task.result()
            case [task] if task == cancel_task:
                catch_task.cancel()  # Stop command catching
                raise commons.CommandAbortedError()
            case []:  # Exception in one of the coroutines
                catch_task.cancel()
                cancel_task.cancel()
                raise commons.CommandAbortedError()
            case _:
                raise RuntimeError(f"heu -> {done}")

    @classmethod
    def _journey_command(cls, callable):
        @functools.wraps(callable)
        async def new_callable(interaction: discord.Interaction, **kwargs):
            async def _run_command(_journey: DiscordJourney):  # Closure
                await callable(_journey, **kwargs)

            if (
                cls._catch_next_command_channel
                and interaction.channel == cls._catch_next_command_channel
                and interaction.user == cls._catch_next_command_member
            ):
                cls._catch_next_command_channel = None
                cls._catch_next_command_member = None
                cls._catch_next_command_coro.send((interaction, _run_command))
                return  # Do not execute command

            async with cls(interaction) as journey:
                await _run_command(journey)

        new_callable._callable = callable

        return new_callable


def journey_command(callable: Callable[[DiscordJourney], None]) -> Callable[[discord.Interaction], None]:
    """Decorator: add the Journey interface to a slash command.

    Replaces the ``interaction: discord.Interaction`` in the callable signature by ``journey: DiscordJourney``.
    """
    return DiscordJourney._journey_command(callable)


_CT = TypeVar("_CT", discord.Message, discord.Member)


def journey_context_menu(callable: Callable[[DiscordJourney, _CT], None]) -> Callable[[discord.Interaction, _CT], None]:
    """Decorator: add the Journey interface to a context menu command.

    Replaces the ``interaction: discord.Interaction`` in the callable signature by ``journey: DiscordJourney``.
    """

    @functools.wraps(callable)
    async def new_callable(interaction: discord.Interaction, arg: _CT):
        async with DiscordJourney(interaction, ephemeral=True) as journey:
            await callable(journey, arg)

    return new_callable



from discord import app_commands


class UserInputError(app_commands.AppCommandError):
    def __init__(self, param: str, message: str, *args: object) -> None:
        self.param = param
        self.message = message
        super().__init__(*args)


class CommandAbortedError(app_commands.AppCommandError):
    def __init__(self, msg: str | None = None) -> None:
        super().__init__(msg or "Mission aborted.")



"""lg-rez / bdd / Modèle de données

Déclaration de toutes les tables et leurs colonnes

"""

from __future__ import annotations

import sqlalchemy

from lgrez.bdd import base
from lgrez.bdd.base import autodoc_Column, autodoc_ManyToOne, autodoc_OneToMany


# Tables de données


class Reaction(base.TableBase):
    """Table de données des réactions d'IA connues du bot.

    Les instances sont enregistrées via :meth:`\/addIA
    <.IA.GestionIA.GestionIA.addIA.callback>` et supprimées via
    :meth:`\/modifIA <.IA.GestionIA.GestionIA.modifIA.callback>`.
    """

    id: int = autodoc_Column(
        sqlalchemy.Integer(),
        primary_key=True,
        doc="Identifiant unique de la réaction, sans signification",
    )
    reponse: str = autodoc_Column(
        sqlalchemy.String(2000),
        nullable=False,
        doc="Réponse, suivant le format (mini-langage) personnalisé " '(``"txt <||> txt <&&> <##>react"``)',
    )

    # One-to-manys
    triggers: list[Trigger] = autodoc_OneToMany(
        "Trigger",
        back_populates="reaction",
        doc="Déclencheurs associés",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        extract = self.reponse.replace("\n", " ")[:15] + "..."
        return f"<Reaction #{self.id} ({extract})>"


class Trigger(base.TableBase):
    """Table de données des mots et expressions déclenchant l'IA du bot.

    Les instances sont enregistrées via :meth:`\/addIA
    <.IA.GestionIA.GestionIA.addIA.callback>` ou :meth:`\/modifIA
    <.IA.GestionIA.GestionIA.modifIA.callback>` et supprimées via
    :meth:`\/modifIA <.IA.GestionIA.GestionIA.modifIA.callback>`.
    """

    id: int = autodoc_Column(
        sqlalchemy.Integer(),
        primary_key=True,
        doc="Identifiant unique du déclencheur, sans signification",
    )
    trigger: str = autodoc_Column(
        sqlalchemy.String(500),
        nullable=False,
        doc="Mots-clés / expressions",
    )

    _reac_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("reactions.id"),
        nullable=False,
    )
    reaction: Reaction = autodoc_ManyToOne(
        "Reaction",
        back_populates="triggers",
        doc="Réaction associée",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        extract = self.trigger.replace("\n", " ")[:15] + "..."
        return f"<Trigger #{self.id} ({extract})>"



"""lg-rez / blocs / Vérification de structure

"""

from contextlib import contextmanager
import enum
import types
from typing import Any, Iterator, Iterable

import discord


class Union(list):
    """Represent a union of possible structures.

    Args:
        *args: Accepted structures.
    """

    def __init__(self, *args) -> None:
        super().__init__(args)  # Pack items

    def __hash__(self) -> int:
        return 0  # don't do that

    def __str__(self) -> str:
        return "  OR  ".join(str(item) for item in self)

    def __repr__(self) -> str:
        return "  OR  ".join(repr(item) for item in self)


Data = dict | list | str | int | float | bool | None
Model = Union | type | types.UnionType | dict | list | None


class MatchError(RuntimeError):
    """Error occuring when matching a structure.

    Derives from :exc:`RuntimeError`.

    Args:
        msg: Exception message.
        data: Object we try to match.
        model: Structure we want to match on.
        path: JSON-like path from the structure root.
    """

    def __init__(
        self,
        msg: str,
        data: Data,
        model: Model,
        path: list[str],
    ) -> None:
        path_str = " > ".join(path)
        lines = [
            msg,
            f"  * Path: {path_str}",
            f"  * Model: {model!r}",
            f"  * Data: {data!r}",
        ]
        super().__init__("\n".join(lines))


class _StructureMatcher:
    def __init__(self) -> None:
        self.path = []

    @contextmanager
    def _dig_in(self, step: str) -> Iterator[None]:
        self.path.append(step)
        try:
            yield
        finally:
            self.path.pop(-1)

    def _check_structure_generic_dict(self, data: dict, model: dict) -> None:
        model_key, model_value = next(iter(model.items()))
        for key, value in data.items():
            with self._dig_in("(new key)"):
                self.check_structure(key, model_key)
            with self._dig_in(f'"{key}"'):
                self.check_structure(value, model_value)

    def _check_structure_dict(self, data: dict, model: dict) -> None:
        for key, model_value in model.items():
            if key not in data:
                raise MatchError(f"Missing {key!r} in data.", data, model, self.path)
            with self._dig_in(key):
                self.check_structure(data[key], model_value)

    def _check_structure_generic_list(self, data: list, model: list) -> None:
        model_item = next(iter(model))
        for count, item in enumerate(data):
            with self._dig_in(f"[{count}]"):
                self.check_structure(item, model_item)

    def _check_structure_list(self, data: list, model: list) -> None:
        if len(data) != len(model):
            raise MatchError("Lists are not of same length.", data, model, self.path)
        for count, (data_item, model_item) in enumerate(zip(data, model)):
            with self._dig_in(f"[{count}]"):
                self.check_structure(data_item, model_item)

    def _check_structure_enum(self, data: Any, model: enum.EnumMeta) -> None:
        try:
            model[data]
        except (KeyError, TypeError):
            raise MatchError(
                f"Data is not one in the values list.",
                data,
                Union(elem.name for elem in model),
                self.path,
            ) from None

    @staticmethod
    def _is_generic(iterable: list | dict) -> bool:
        if len(iterable) != 1:
            return False
        key = next(iter(iterable))
        if isinstance(key, str | int | float | bool | type(None)):
            return False
        return True

    def test_structure(self, data: Data, model: Model) -> bool:
        try:
            self.check_structure(data, model)
        except MatchError:
            return False
        else:
            return True

    def check_structure(self, data: Data, model: Model) -> None:
        match model, data:
            # OR
            case Union(), _:
                for submodel in model:
                    if self.test_structure(data, submodel):
                        break
                else:
                    raise MatchError("Could not match data on either of models.", data, model, self.path)

            # Type restriction
            case (enum.EnumMeta() | discord.enums.EnumMeta()), _:
                self._check_structure_enum(data, model)
            case type(), model():  # isinstance but 500 IQ (self-slurping yes)
                pass
            case type(), _:
                raise MatchError(f"Data is not of requested type.", data, model, self.path)
            case types.UnionType(), _:
                if not isinstance(data, model):
                    raise MatchError(f"Data is not of any of the requested types.", data, model, self.path)

            # Submodel
            case dict(), dict() if self._is_generic(model):
                self._check_structure_generic_dict(data, model)
            case dict(), dict():
                self._check_structure_dict(data, model)
            case list(), list() if self._is_generic(model):
                self._check_structure_generic_list(data, model)
            case list(), list():
                self._check_structure_list(data, model)
            case (dict() | list()), _:
                raise MatchError(f"Data is not of requested type.", data, model, self.path)

            # Literal
            case (str() | int() | float() | True | False | None), _:
                if data != model:
                    raise MatchError(f"Data is not the required literal.", data, model, self.path)

            # Fallback
            case _:
                raise MatchError(f"Could not match.", data, model, self.path)


def check_structure(data: Data, model: Model) -> None:
    """Check that a given object matches a required structure.

    Args:
        data: Object to match.
        model: Structure to match on.

    Raises:
        .MatchError: If matching failed.
    """
    _StructureMatcher().check_structure(data, model)


def check_server_structure(
    structure: dict,
    required_roles: Iterable[str],
    required_channels: Iterable[str],
    required_emojis: Iterable[str],
) -> None:
    """Check that the given dict represents a valid server structure.

    Args:
        structure: The dict to check.
        required_roles: The slugs of roles needed in the structure.
        required_channels: The slugs of channels needed in the structure.
        required_emojis: The slugs of emojis needed in the structure.

    Raises:
        .MatchError: If matching failed.
    """
    # Pseudo-types autorisés dans la structure
    role = Union("@everyone", *structure.get("roles", {}))
    channel = Union()
    for categ in structure.get("categories", {}).values():
        channel.extend(categ.get("channels", {}))
    permission = Union(*discord.Permissions.VALID_FLAGS)
    PermissionsGroup = enum.Enum(
        "PermissionsGroup",
        {
            prop: getattr(discord.Permissions, prop)()
            for prop in [
                "advanced",
                "all",
                "all_channel",
                "general",
                "membership",
                "stage",
                "stage_moderator",
                "text",
                "voice",
            ]
        },
    )

    # Structure requise
    required_structure = {
        "name": str,
        "icon": Union(
            None,
            {
                "drive": bool,
                "png_path_or_id": str,
            },
        ),
        "afk_channel": Union(channel, None),
        "afk_timeout": int,
        "verification_level": discord.VerificationLevel,
        "default_notifications": discord.NotificationLevel,
        "explicit_content_filter": discord.ContentFilter,
        "system_channel": channel,
        "system_channel_flags": {
            "join_notifications": bool,
            "premium_subscriptions": bool,
        },
        "preferred_locale": str,
        "roles": {
            str: {
                "name": str,
                "color": str,
                "hoist": bool,
                "mentionable": bool,
                "permissions": Union(PermissionsGroup, [permission]),
            },
        },
        "base_role": role,
        "everyone_permissions": [permission],
        "categories": {
            str: {
                "name": str,
                "overwrites": {role: {permission: bool}},
                "channels": {
                    str: {
                        "name": str,
                        "topic": str | None,
                        "overwrites": {role: {permission: bool | None}},
                    },
                },
                "voice_channels": {
                    str: {
                        "name": str,
                        "overwrites": {role: {permission: bool | None}},
                    },
                },
            },
        },
        "emojis": {
            "drive": bool,
            "folder_path_or_id": str,
            "required": {str: str},
            "restrict_roles": {str: Union(None, [role])},
        },
    }

    check_structure(structure, required_structure)

    # Vérification éléments imposés minimums
    for role in required_roles:
        if role == "everyone":
            continue
        if not role in structure["roles"]:
            raise RuntimeError('config.server_structure["roles"]: ' f'Missing role "{role}", required for playing!')
    for channel in required_channels:
        for categ in structure["categories"].values():
            if channel in categ["channels"]:
                break
        else:
            raise RuntimeError(
                'config.server_structure["categories"][<any categ>]'
                '["channels"]: '
                f'Missing channel "{channel}", required for playing!'
            )
    for emoji in required_emojis:
        if not emoji in structure["emojis"]["required"]:
            raise RuntimeError('config.server_structure["emojis"]: ' f'Missing emoji "{emoji}", required for playing!')



"""lg-rez / features / IA des réponses

Tout ce qui concerne la manière dont le bot réagit aux messages :
détermination de la meilleure réaction, gestion des réactions,
activation/désactivation des modes de chat

"""

import re
import random
from typing import Callable, Coroutine, Literal
import requests

import discord
from discord import app_commands

from lgrez import config
from lgrez.blocs import tools
from lgrez.blocs.journey import DiscordJourney, journey_command, journey_context_menu
from lgrez.bdd import Trigger, Reaction, Role


# Marqueurs de séparation du mini-langage des séquences-réactions
MARK_OR = " <||> "
MARK_THEN = " <&&> "
MARK_REACT = "<::>"
MARKS = [MARK_OR, MARK_THEN, MARK_REACT]


async def _build_sequence(journey: DiscordJourney) -> str:
    """Construction d'une séquence-réaction par l'utilisateur"""
    reponse = ""
    channel_id = journey.channel.id
    try:
        if channel_id not in config.bot.in_stfu:
            config.bot.in_stfu.append(channel_id)

        while True:
            message = await journey.channel.send(
                "Réaction du bot : prochain message/commande/média, ou réaction à ce message"
            )
            ret = await tools.wait_for_react_added(message, process_text=True)
            if isinstance(ret, str):
                reponse += ret
            else:  # React
                reponse += MARK_REACT + ret.name

            choix = await journey.buttons(
                "Et ensuite ?",
                {
                    MARK_THEN: discord.ui.Button(label="Puis", emoji="▶"),
                    MARK_OR: discord.ui.Button(label="Alternatif", emoji="🔀"),
                    False: discord.ui.Button(label="Fin", emoji="⏹"),
                },
            )

            if choix:
                # On ajoute la marque OR ou THEN à la séquence
                reponse += choix
            else:
                break

    finally:
        if channel_id in config.bot.in_stfu:
            config.bot.in_stfu.remove(channel_id)

    return reponse


def fetch_tenor(trigger):
    """Renvoie le GIF Tenor le plus "pertinent" pour un texte donné.

    Args:
        trigger (str): texte auquel réagir.

    Returns:
        :class:`str` (URL du GIF) ou ``None``
    """
    # API key module ternorpy (parce que la flemme de créer un compte Tenor)
    apikey = "J5UVWPVIM4A5"

    rep = requests.get(
        url="https://api.tenor.com/v1/search",
        params={
            "q": trigger,
            "key": apikey,
            "limit": 1,
            "locale": "fr_FR",
            "contentfilter": "off",
            "media_filter": "minimal",
            "ar_range": "all",
        },
    )

    if rep:
        gifs = rep.json()["results"]
        # Payload Tenor : {..., "results":[ ... ]}
        # (https://tenor.com/gifapi/documentation#responseobjects-gif)
        if gifs:
            return gifs[0]["itemurl"]

    return None  # Pas de GIF trouvé


DESCRIPTION = """Commandes relatives à l'IA (réponses automatiques du bot)"""


@app_commands.command()
@tools.private()
@journey_command
async def stfu(journey: DiscordJourney, *, force: Literal["on", "off"] = None):
    """Active/désactive la réponse automatique du bot sur ton channel privé.

    Args:
        force: Forcer l'activation / la désactivation (toggle par défaut).

    N'agit que sur les messages classiques envoyés dans le channel : les commandes restent reconnues.

    Si vous ne comprenez pas le nom de la commande, demandez à Google.
    """
    id = journey.channel.id

    if force in [None, "on"] and id not in config.bot.in_stfu:
        config.bot.in_stfu.append(id)
        await journey.send("Okay, je me tais ! Tape `/stfu` quand tu voudras de nouveau de moi :cry:")

    elif force in [None, "off"] and id in config.bot.in_stfu:
        config.bot.in_stfu.remove(id)
        await journey.send("Ahhh, ça fait plaisir de pouvoir reparler !")


@app_commands.command()
@journey_command
async def fals(journey: DiscordJourney, force: Literal["on", "off"] = None):
    """Active/désactive le mode « foire à la saucisse ».

    Args:
        force: Forcer l'activation / la désactivation (toggle par défaut).

    En mode « foire à la saucisse », le bot réagira à (presque) tous les messages,
    pas seulement sur les motifs qu'on lui a appris.

    À utiliser à vos risques et périls !
    """
    id = journey.channel.id

    if force in [None, "on"] and id not in config.bot.in_fals:
        config.bot.in_fals.append(id)
        await journey.send("https://tenor.com/view/saucisse-sausage-gif-5426973")

    elif force in [None, "off"] and id in config.bot.in_fals:
        config.bot.in_fals.remove(id)
        await journey.send("T'as raison, faut pas abuser des bonnes choses")


class _FakeMessage:
    def __init__(self, channel: discord.TextChannel, author: discord.Member, content: str) -> None:
        self.channel = channel
        self.author = author
        self.content = content

    async def add_reaction(emoji: discord.Emoji | str) -> None:
        pass


@app_commands.command()
@journey_command
async def react(journey: DiscordJourney, *, trigger: str):
    """Force le bot à réagir à un message (sur un chan public, en mode STFU...)

    Args:
        trigger: Texte auquel le bot doit réagir.

    Permet de faire appel à l'IA du bot même sur les chans publics,ou en mode STFU, etc.

    Si utilisée par un MJ, active aussi le mode débug des évaluations Python (messages d'erreur).
    """
    message = _FakeMessage(journey.channel, journey.member, trigger)
    debug = journey.member.top_role >= config.Role.mj
    await process_ia(message, journey.send, debug=debug)


@app_commands.command()
@journey_command
async def reactfals(journey: DiscordJourney, *, trigger: str):
    """Force le bot à réagir à un message comme en mode Foire à la saucisse.

    Args:
        trigger: Texte auquel le bot doit réagir.

    Permet de faire appel directement au mode Foire à la saucisse,
    même si il n'est pas activé / sur un chan public.
    """
    gif = fetch_tenor(trigger)
    await journey.send(gif or "Palaref")


async def _add_ia(journey: DiscordJourney, *, triggers: str, reponse: str | None):
    triggers = triggers.split(";")
    triggers = [tools.remove_accents(s).lower().strip() for s in triggers]
    triggers = list({trig for trig in triggers if trig})
    # filtre doublons (accents et non accents...) et triggers vides

    avoided = []
    for trigger in triggers.copy():
        if Trigger.query.filter_by(trigger=trigger).all():
            avoided.append(trigger)
            triggers.remove(trigger)

    avoided = f"Trigger(s) déjà associé(s) à une réaction et ignorés : {avoided}\n\n" if avoided else ""

    if not triggers:
        await journey.send(avoided + ":x: Aucun trigger valide, abort")
        return

    await journey.send(avoided + f":arrow_forward: Triggers : `{'` – `'.join(triggers)}`")

    if reponse:
        reponse = reponse.strip()
    else:
        reponse = await _build_sequence(journey)

    if not reponse:
        await journey.channel.send("Réponse textuelle vide interdite, abort.")
        return

    await journey.channel.send(f"Résumé de la séquence : {tools.code(reponse)}")
    async with journey.channel.typing():
        reac = Reaction(reponse=reponse)
        config.session.add(reac)

        trigs = [Trigger(trigger=trigger, reaction=reac) for trigger in triggers]
        config.session.add_all(trigs)

        config.session.commit()

    await journey.channel.send("Règle ajoutée en base.")


@app_commands.command()
@tools.mjs_et_redacteurs
@journey_command
async def add_ia(journey: DiscordJourney, *, triggers: str, reponse: str | None = None):
    """Ajoute une règle d'IA (COMMANDE MJ/RÉDACTEURS)

    Args:
        triggers: Mot(s), phrase(s) ou expression(s) séparées par des ";".
        reponse: Si réponse textuelle simple, pour ajout rapide.

    Une sécurité empêche d'ajouter un trigger déjà existant.

    Dans le cas où plusieurs expressions sont spécifiées, toutes déclencheront l'action demandée.
    """
    await _add_ia(journey, triggers=triggers, reponse=reponse)


@app_commands.context_menu(name="Nouvelle règle d'IA")
@tools.mjs_et_redacteurs
@journey_context_menu
async def add_ia_menu(journey: DiscordJourney, message: discord.Message):
    await _add_ia(journey, triggers=message.content, reponse=None)


@app_commands.command()
@tools.mjs_et_redacteurs
@journey_command
async def list_ia(journey: DiscordJourney, trigger: str | None = None, sensi: float = 0.5):
    """Liste les règles d'IA reconnues par le bot (COMMANDE MJ/RÉDACTEURS)

    Args
        trigger: Mot/expression permettant de filter et trier les résultats.
        sensi: Sensibilité de détection (ratio des caractères correspondants, entre 0 et 1), défaut 0.5.
    """
    if trigger:
        trigs = Trigger.find_nearest(trigger, col=Trigger.trigger, sensi=sensi, solo_si_parfait=False)
        if not trigs:
            await journey.send(f"Rien trouvé, pas de chance (sensi = {sensi})")
            return
    else:
        raw_trigs = Trigger.query.order_by(Trigger.id).all()
        # Trié par date de création
        trigs = list(zip(raw_trigs, [None] * len(raw_trigs)))
        # Mise au format (trig, score)

    reacts = []  # Réactions associées à notre liste de triggers
    for trig in trigs:
        if (reac := trig[0].reaction) not in reacts:
            # Pas de doublons, et reste ordonné
            reacts.append(reac)

    def nettoy(s):
        # Abrège la réponse si trop longue et neutralise les
        # sauts de ligne / rupture code_bloc, pour affichage
        s = s.replace("\r\n", "\\n").replace("\n", "\\n")
        s = s.replace("\r", "\\r").replace("```", "'''")
        if len(s) < 75:
            return s
        else:
            return s[:50] + " [...] " + s[-15:]

    rep = ""
    for reac in reacts:  # pour chaque réponse
        r = ""
        for (trig, score) in trigs:  # pour chaque trigger
            if trig.reaction == reac:
                sc = f"({float(score):.2}) " if score else ""
                r += f" - {sc}{trig.trigger}"
                # (score) trigger - (score) trigger ...

        rep += r.ljust(50) + f" ⇒ {nettoy(reac.reponse)}\n"
        # ⇒ réponse

    rep += "\nPour modifier une réaction, utiliser `/modif_ia <trigger>`."

    await journey.send(rep, code=True)


@app_commands.command()
@tools.mjs_et_redacteurs
@journey_command
async def modif_ia(journey: DiscordJourney, *, trigger: str):
    """Modifie/supprime une règle d'IA (COMMANDE MJ/RÉDACTEURS)

    Args:
        trigger: Mot/expression déclenchant la réaction à modifier/supprimer.

    Permet d'ajouter et supprimer des triggers, de modifier la réaction du bot
    (construction d'une  séquence de réponses successives ou aléatoires)
    ou de supprimer la réaction.
    """
    trigs = Trigger.find_nearest(trigger, col=Trigger.trigger)
    if not trigs:
        await journey.send("Rien trouvé.")
        return

    trig = trigs[0][0]
    reac = trig.reaction

    displ_seq = reac.reponse if reac.reponse.startswith("`") else tools.code(reac.reponse)  # Pour affichage
    trigs = list(reac.triggers)

    choix = await journey.buttons(
        f"Triggers : `{'` – `'.join([trig.trigger for trig in trigs])}`\nSéquence réponse : {displ_seq}\n\nModifier :",
        {
            "triggers": discord.ui.Button(label="Triggers", emoji="⏩"),
            "response": discord.ui.Button(label="Réponse", emoji="⏺"),
            "delete": discord.ui.Button(label="Supprimer", emoji="🚮"),
        },
    )

    if choix == "triggers":  # Modification des triggers
        while True:
            action = await journey.buttons(
                "Supprimer / ajouter un trigger :",
                {
                    index: discord.ui.Button(label=trigger.trigger, emoji="🚮", style=discord.ButtonStyle.danger)
                    for index, trigger in enumerate(trigs)
                }
                | {
                    "new": discord.ui.Button(label="Nouveau", emoji="🆕", style=discord.ButtonStyle.primary),
                    "stop": discord.ui.Button(label="Fini", emoji="⏹"),
                },
            )

            if action == "stop":
                break
            elif action == "new":
                (new_trigger,) = await journey.modal("Ajouter un trigger", "Nouveau déclencheur")
                new_trig = Trigger(trigger=new_trigger, reaction=reac)
                trigs.append(new_trig)
                config.session.add(new_trig)
                config.session.commit()
            else:
                config.session.delete(trigs.pop(action))
                config.session.commit()

        if not trigs:  # on a tout supprimé !
            await journey.send("Tous les triggers supprimés, suppression de la réaction")
            config.session.delete(reac)
            config.session.commit()
            return

    elif choix == "response":  # Modification de la réponse
        if any([mark in reac.reponse for mark in MARKS]):
            # Séquence compliquée
            await journey.send(
                "\nLa séquence-réponse peut être refaite manuellement "
                "ou modifiée rapidement en envoyant directement la "
                "séquence ci-dessus modifiée (avec les marqueurs : "
                f"OU = {tools.code(MARK_OR)}, "
                f"ET = {tools.code(MARK_THEN)}, "
                f"REACT = {tools.code(MARK_REACT)}"
            )

        reponse = await _build_sequence(journey)
        if not reponse:
            await journey.send("Réponse textuelle vide interdite, abort.")

        reac.reponse = reponse

    else:  # Suppression
        config.session.delete(reac)
        for trig in trigs:
            config.session.delete(trig)

    config.session.commit()

    await journey.send("Fini.")


async def trigger_at_mj(message: discord.Message, send_callable: Callable[[str], Coroutine]) -> bool:
    """Règle d'IA : réaction si le message mentionne les MJs.

    Args:
        message: Message auquel réagir.

    Returns:
        Si le message mentionne les MJ et qu'une réponse a été envoyée
    """
    if config.Role.mj.mention in message.content:
        await send_callable("Les MJs ont entendu ton appel, ils sont en route ! :superhero:")
        return True

    return False


async def trigger_roles(
    message: discord.Message, send_callable: Callable[[str], Coroutine], sensi: float = 0.8
) -> bool:
    """Règle d'IA : réaction si un nom de rôle est donné.

    Args:
        message: Message auquel réagir.
        sensi: Sensibilité de la recherche (voir :meth:`.bdd.base.TableMeta.find_nearest`).

    Trouve l'entrée la plus proche de ``message.content`` dans la table :class:`.bdd.Role`.

    Returns:
        Si un rôle a été trouvé (sensibilité ``> sensi``) et qu'une réponse a été envoyée.
    """
    roles = Role.find_nearest(message.content, col=Role.nom, filtre=(Role.actif.is_(True)), sensi=sensi)

    if roles:  # Au moins un trigger trouvé à cette sensi
        await send_callable(embed=roles[0][0].embed)
        return True

    return False


async def trigger_reactions(
    message: discord.Message,
    send_callable: Callable[[str], Coroutine],
    chain: str | None = None,
    sensi: float = 0.7,
    debug: bool = False,
) -> bool:
    """Règle d'IA : réaction à partir de la table :class:`.bdd.Reaction`.

    Args:
        message: Message auquel réagir.
        chain: Contenu auquel réagir (défaut : contenu de ``message``).
        sensi: Sensibilité de la recherche (cf :meth:`.bdd.base.TableMeta.find_nearest`).
        debug: Si ``True``, affiche les erreurs lors de l'évaluation des messages (voir :func:`.tools.eval_accols`).

    Trouve l'entrée la plus proche de ``chain`` dans la table :class:`.bdd.Reaction` ;
    si il contient des accolades, évalue le message selon le contexte de ``message``.

    Returns:
        Si une réaction a été trouvé (sensibilité ``> sensi``) et qu'une réponse a été envoyée.
    """
    if not chain:  # Si pas précisé,
        chain = message.content  # contenu de message
    trigs = Trigger.find_nearest(chain, col=Trigger.trigger, sensi=sensi)

    if trigs:  # Au moins un trigger trouvé à cette sensi
        trig = trigs[0][0]  # Meilleur trigger (score max)
        seq = trig.reaction.reponse  # Séquence-réponse associée

        for rep in seq.split(MARK_THEN):  # Pour chaque étape :
            if MARK_OR in rep:
                # Si plusieurs possibilités, on en choisit une random
                rep = random.choice(rep.split(MARK_OR))

            if rep.startswith(MARK_REACT):  # Réaction
                react = rep.lstrip(MARK_REACT)
                emoji = tools.emoji(react, must_be_found=False) or react
                await message.add_reaction(emoji)

            else:  # Sinon, texte / média
                # On remplace tous les "{expr}" par leur évaluation
                rep = tools.eval_accols(rep, locals_=locals(), debug=debug)
                await send_callable(rep)

        return True

    return False


async def trigger_sub_reactions(
    message: discord.Message, send_callable: Callable[[str], Coroutine], sensi: float = 0.9, debug: bool = False
) -> bool:
    """Règle d'IA : réaction à partir de la table, mais sur les mots

    Appelle :func:`trigger_reactions(bot, message, mot, sensi, debug) <.trigger_reactions>`
    pour tous les mots ``mot`` composant ``message.content`` (mots de plus de 4 lettres,
    testés des plus longs aux plus courts).

    Args:
        message: Message auquel réagir.
        sensi: Sensibilité de la recherche (cf :meth:`.bdd.base.TableMeta.find_nearest`).
        debug: Si ``True``, affiche les erreurs lors de l'évaluation des messages (voir :func:`.tools.eval_accols`).

    Returns:
        Si une réaction a été trouvé (sensibilité ``> sensi``) et qu'une réponse a été envoyée.
    """
    mots = message.content.split(" ")
    if len(mots) > 1:  # Si le message fait plus d'un mot
        for mot in sorted(mots, key=lambda m: -len(m)):
            # On parcourt les mots du plus long au plus court
            if len(mot) > 4:  # on élimine les mots de liaison
                if await trigger_reactions(message, send_callable, chain=mot, sensi=sensi, debug=debug):
                    # Si on trouve une sous-rect (à 0.9)
                    return True

    return False


async def trigger_di(message: discord.Message, send_callable: Callable[[str], Coroutine]) -> bool:
    """Règle d'IA : réaction aux messages en di... / cri...

    Args:
        message: Message auquel réagir.

    Returns:
        Si le message correspond et qu'une réponse a été envoyée.
    """
    c = message.content
    diprefs = ["di", "dy", "dis ", "dit ", "dis-", "dit-"]
    criprefs = ["cri", "cry", "kri", "kry"]
    pos_prefs = {c.lower().find(pref): pref for pref in diprefs + criprefs if pref in c[:-1].lower()}
    # On extrait les cas où le préfixe est à la fin du message

    if pos_prefs:  # Si on a trouvé au moins un préfixe
        i = min(pos_prefs)
        pref = pos_prefs[i]
        if pref in criprefs:
            mess = tools.bold(c[i + len(pref) :].upper())
        else:
            mess = c[i + len(pref) :]
        await send_callable(mess, tts=True)
        # On envoie le di.../cri... en mode TTS (oh si, c'est rigolo)
        return True

    return False


async def trigger_gif(message: discord.Message, send_callable: Callable[[str], Coroutine]) -> bool:
    """Règle d'IA : réaction par GIF en mode Foire à la saucisse.

    Args:
        message: Message auquel réagir.

    Returns:
        Si le message correspond et qu'une réponse a été envoyée.
    """
    if message.channel.id in config.bot.in_fals:
        # Chan en mode Foire à la saucisse
        async with message.channel.typing():
            gif = fetch_tenor(message.content)
            if gif:
                await send_callable(gif)
                return True

    return False


async def trigger_mot_unique(message: discord.Message, send_callable: Callable[[str], Coroutine]) -> bool:
    """Règle d'IA : réaction à un mot unique (le répète).

    Args:
        message: Message auquel réagir.

    Returns:
        Si le message correspond et qu'une réponse a été envoyée
    """
    if len(message.content.split()) == 1 and ":" not in message.content:
        # : pour ne pas trigger aux liens
        rep = f"{message.content.capitalize()} ?"
        await send_callable(rep)
        return True

    return False


async def trigger_a_ou_b(message: discord.Message, send_callable: Callable[[str], Coroutine]) -> bool:
    """Règle d'IA : réaction à un motif type « a ou b » (répond « b »).

    Args:
        message: Message auquel réagir.

    Returns:
        Si le message correspond et qu'une réponse a été envoyée.
    """
    if motif := re.fullmatch(r"(.+)\s+ou\s+(.+?)", message.content):
        rep = f"{motif.group(2).rstrip(' !?.,;')}.".capitalize()
        await send_callable(rep)
        return True

    return False


async def default(message: discord.Message, send_callable: Callable[[str], Coroutine]) -> bool:
    """Règle d'IA : réponse par défaut

    Args:
        message: Message auquel réagir.

    Returns:
        Si le message correspond et qu'une réponse a été envoyée.
    """
    mess = "Désolé, je n'ai pas compris :person_shrugging:"
    if random.random() < 0.05:
        mess += "\n(et toi, tu as perdu)"
    await send_callable(mess)
    return True


async def process_ia(message: discord.Message, send_callable: Callable[[str], Coroutine], debug: bool = False) -> None:
    """Exécute les règles d'IA.

    Args:
        message: Message auquel réagir.
        debug: Si ``True``, affiche les erreurs lors de l'évaluation des messages (voir :func:`.tools.eval_accols`).
    """
    (
        await trigger_at_mj(message, send_callable)  # @MJ (aled)
        or await trigger_gif(message, send_callable)  # Un petit GIF ? (si FALS)
        or await trigger_roles(message, send_callable)  # Rôles
        or await trigger_reactions(message, send_callable, debug=debug)  # Table Reaction ("IA")
        or await trigger_sub_reactions(message, send_callable, debug=debug)  # IA sur les mots
        or await trigger_a_ou_b(message, send_callable)  # "a ou b" ==> "b"
        or await trigger_di(message, send_callable)  # di... / cri...
        or await trigger_mot_unique(message, send_callable)  # Un seul mot ==> on répète
        or await default(message, send_callable)  # Réponse par défaut
    )



"""lg-rez / Blocs transversaux

Chaque module de ``lgrez.blocs`` réalise une tâche non liée à une
fonctionnalité spécifique du bot : connection à un service, outils
divers...

"""

import os

dir = os.path.dirname(os.path.realpath(__file__))

__all__ = []
for file in os.listdir(dir):
    if not file.endswith(".py"):
        # Not a Python file
        continue

    name = file[:-3]
    if name.startswith("_"):
        # Private / magic module
        continue

    # Public submodule: add to __all__
    __all__.append(name)



"""lg-rez / features / Actions publiques

Gestion des haros, candidatures à la mairie, résultats des votes

"""

from typing import Literal

import discord
from discord import app_commands

from lgrez import config
from lgrez.blocs import tools
from lgrez.blocs.journey import DiscordJourney, journey_command, journey_context_menu
from lgrez.bdd import Joueur, CandidHaro, CandidHaroType, Vote
from lgrez.features.voter_agir import do_vote


DESCRIPTION = """Commandes d'actions vous engageant publiquement"""


async def _haro(journey: DiscordJourney, joueur: Joueur):
    moi = Joueur.from_member(journey.member)
    try:
        vaction = joueur.action_vote(Vote.cond)
    except RuntimeError:
        await journey.send(":x: Minute papillon, le jeu n'est pas encore lancé !")
        return

    if not vaction.is_open:
        await journey.send(":x: Pas de vote pour le condamné du jour en cours !")
        return

    (motif,) = await journey.modal(
        f"Haro contre {joueur.nom}",
        discord.ui.TextInput(label="Quelle est la raison de cette haine ?", style=discord.TextStyle.paragraph),
    )

    emb = discord.Embed(
        title=(f"**{config.Emoji.ha}{config.Emoji.ro} contre {joueur.nom} !**"),
        description=f"**« {motif} »\n**",
        color=0xFF0000,
    )
    emb.set_author(name=f"{moi.nom} en a gros 😡😡")
    emb.set_thumbnail(url=config.Emoji.bucher.url)

    await journey.ok_cancel("C'est tout bon ?", embed=emb)

    class _HaroView(discord.ui.View):
        @discord.ui.button(
            label=f"Voter contre {joueur.nom}"[:80], style=discord.ButtonStyle.primary, emoji=config.Emoji.bucher
        )
        async def vote(self, vote_interaction: discord.Interaction, button: discord.ui.Button):
            async with DiscordJourney(vote_interaction, ephemeral=True) as vote_journey:
                try:
                    votant = Joueur.from_member(vote_journey.member)
                except ValueError:
                    await vote_journey.send(":x: Tu n'as pas le droit de vote, toi")
                await do_vote(vote_journey, Vote.cond, votant=votant, cible=joueur)

        @discord.ui.button(label=f"Contre-haro", style=discord.ButtonStyle.danger, emoji=config.Emoji.ha)
        async def contre_haro(self, contre_haro_interaction: discord.Interaction, button: discord.ui.Button):
            async with DiscordJourney(contre_haro_interaction, ephemeral=True) as contre_haro_journey:
                await _haro(contre_haro_journey, joueur=moi)

    haro_message = await config.Channel.haros.send(f"(Psst, {joueur.member.mention} :3)", embed=emb, view=_HaroView())
    await config.Channel.debats.send(
        f"{config.Emoji.ha}{config.Emoji.ro} de {journey.member.mention} sur {joueur.member.mention} ! "
        f"Vous en pensez quoi vous ? (détails sur {config.Channel.haros.mention})"
    )

    haro = CandidHaro(joueur=joueur, type=CandidHaroType.haro, message_id=haro_message.id)
    contre_haro = CandidHaro(joueur=moi, type=CandidHaroType.haro)
    CandidHaro.add(haro, contre_haro)

    if journey.channel != config.Channel.haros:
        await journey.send(f"Allez, c'est parti ! ({config.Channel.haros.mention})")


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def haro(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.VivantTransformer]):
    """Lance publiquement un haro contre une autre personne.

    Args:
        joueur: Le joueur ou la joueuse à accuser de tous les maux.

    Cette commande n'est utilisable que lorsqu'un vote pour le condamné est en cours.
    """
    await _haro(journey, joueur=joueur)


@app_commands.context_menu(name="Lancer un haro contre ce joueur")
@tools.vivants_only
@journey_context_menu
async def haro_menu(journey: DiscordJourney, member: discord.Member):
    if member.top_role >= config.Role.mj:
        await journey.send(":x: Attends deux secondes, tu pensais faire quoi là ?")
        return

    if member == config.bot.user:
        await journey.send(":x: Tu ne peux pas haro le bot, enfin !!!")
        return

    try:
        joueur = Joueur.from_member(member)
    except ValueError:
        await journey.send(":x: Hmm, ce joueur n'a pas l'air inscrit !")
        return

    await _haro(journey, joueur=joueur)


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def candid(journey: DiscordJourney):
    """Déclare ta candidature à l'élection du nouveau maire.

    Cette commande n'est utilisable que lorsqu'un vote pour le nouveau maire est en cours.
    """
    joueur = Joueur.from_member(journey.member)
    try:
        vaction = joueur.action_vote(Vote.cond)
    except RuntimeError:
        await journey.send(":x: Minute papillon, le jeu n'est pas encore lancé !")
        return

    if not vaction.is_open:
        await journey.send(":x: Pas de vote pour le nouveau maire en cours !")
        return

    if CandidHaro.query.filter_by(joueur=joueur, type=CandidHaroType.candidature).first():
        await journey.send(":x: Hola collègue, tout doux, tu t'es déjà présenté(e) !")
        return

    (motif,) = await journey.modal(
        "Candidature à la Mairie",
        discord.ui.TextInput(label="Quel est ton programme politique ?", style=discord.TextStyle.paragraph),
    )

    emb = discord.Embed(
        title=(f"**{config.Emoji.maire} {joueur.nom} candidate à la Mairie !**"),
        description=("Voici son programme politique :\n" + tools.bold(motif)),
        color=0xF1C40F,
    )
    emb.set_author(name=f"{joueur.nom} vous a compris !")
    emb.set_thumbnail(url=config.Emoji.maire.url)

    await journey.ok_cancel("C'est tout bon ?", embed=emb)

    class _CandidView(discord.ui.View):
        @discord.ui.button(
            label=f"Voter pour {joueur.nom}"[:80], style=discord.ButtonStyle.success, emoji=config.Emoji.maire
        )
        async def vote(self, vote_interaction: discord.Interaction, button: discord.ui.Button):
            async with DiscordJourney(vote_interaction, ephemeral=True) as vote_journey:
                try:
                    votant = Joueur.from_member(vote_journey.member)
                except ValueError:
                    await vote_journey.send(":x: Oh, tu n'as pas le droit de vote, toi !")
                await do_vote(vote_journey, Vote.maire, votant=votant, cible=joueur)

    candid_message = await config.Channel.haros.send("Here comes a new challenger!", embed=emb, view=_CandidView())
    await config.Channel.debats.send(
        f"{journey.member.mention} se présente à la Mairie ! Vous en pensez quoi vous ?\n"
        f"(détails sur {config.Channel.haros.mention})"
    )
    await journey.send(f"Allez, c'est parti ! ({config.Channel.haros.mention})")

    ch = CandidHaro(joueur=joueur, type=CandidHaroType.candidature, message_id=candid_message.id)
    CandidHaro.add(ch)


@app_commands.command()
@tools.mjs_only
@journey_command
async def wipe(journey: DiscordJourney, quoi: Literal["haros", "candids"]):
    """Efface les haros / candidatures du jour (COMMANDE MJ)

    Args:
        quoi: Type d'objet à supprimer.
    """
    if quoi == "haros":
        cht = CandidHaroType.haro
    elif quoi == "candids":
        cht = CandidHaroType.candidature
    else:
        await journey.send("Mauvais argument")
        return

    candid_haros: list[CandidHaro] = CandidHaro.query.filter_by(type=cht).all()

    if not candid_haros:
        await journey.send("Rien à faire.")
        return

    for candid_haro in candid_haros:
        await candid_haro.disable_message_buttons()
    CandidHaro.delete(*candid_haros)

    await journey.send("Fait.")
    await tools.log(f"/wipe : {len(candid_haros)} {quoi} supprimés")



"""lg-rez / features / Commandes spéciales

Commandes spéciales (méta-commandes, imitant ou impactant le
déroulement des autres ou le fonctionnement du bot)

"""

import asyncio
import os
import sys
import traceback
from typing import Literal

# Unused imports because useful for /do or /shell globals
import discord
from discord import app_commands

from lgrez import __version__, config, features, blocs, bdd, commands, commons
from lgrez.blocs import gsheets, tools, realshell
from lgrez.bdd import *  # toutes les tables dans globals()
from lgrez.blocs.journey import DiscordJourney, journey_command


class CommandTransformer(app_commands.Transformer):
    @staticmethod
    def _get_commands(interaction: discord.Interaction) -> dict[str, app_commands.Command]:
        if getattr(interaction.namespace, "mode") == "enable":
            return config.bot.tree.disabled_commands
        else:
            return config.bot.tree.enabled_commands

    async def transform(self, interaction: discord.Interaction, value: str) -> app_commands.Command:
        return self._get_commands(interaction)[value]

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return [
            app_commands.Choice(name=name, value=name) for name in self._get_commands(interaction) if current in name
        ][:25]


class SubCommandTransformer(app_commands.Transformer):
    @staticmethod
    def _get_commands(interaction: discord.Interaction) -> dict[str, app_commands.Command]:
        return config.bot.tree.enabled_commands_and_subcommands

    async def transform(self, interaction: discord.Interaction, value: str) -> app_commands.Command:
        return self._get_commands(interaction)[value]

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return [
            app_commands.Choice(name=name, value=name) for name in self._get_commands(interaction) if current in name
        ][:25]


DESCRIPTION = """Commandes spéciales (méta-commandes et expérimentations)"""


@app_commands.command()
@tools.mjs_only
@journey_command
async def panik(journey: DiscordJourney):
    """Tue instantanément le bot, sans confirmation (COMMANDE MJ)

    PAAAAANIK
    """
    await journey.interaction.response.defer()
    sys.exit()


@app_commands.command()
@tools.mjs_only
@journey_command
async def do(journey: DiscordJourney, *, code: str):
    """Exécute du code Python et affiche le résultat (COMMANDE MJ)

    Args:
        code: Instruction Python évaluable (dispo : `journey`, `config`, `blocs`, `features`, `bdd`, `<table>`...)

    Si ``code`` est une coroutine, elle sera awaited (ne pas inclure ``await`` dans ``code``).

    Aussi connue sous le nom de « faille de sécurité », cette commande permet de faire environ
    tout ce qu'on veut sur le bot (y compris le crasher, importer des modules, exécuter des fichiers .py...
    même si c'est un peu compliqué) voire d'impacter le serveur sur lequel le bot tourne si on est motivé.

    À utiliser avec parcimonie donc, et QUE pour du développement/debug !
    """
    await journey.interaction.response.defer(thinking=True)

    class Answer:
        rep = None

    _a = Answer()

    locs = globals()
    locs["journey"] = journey
    locs["_a"] = _a
    try:
        exec(f"_a.rep = {code}", locs)
        if asyncio.iscoroutine(_a.rep):
            _a.rep = await _a.rep
    except Exception:
        _a.rep = traceback.format_exc()
    await journey.send(f">>> {code}\n{_a.rep}", code=True, langage="py")


@app_commands.command()
@tools.mjs_only
@journey_command
async def shell(journey: DiscordJourney):
    """Lance un terminal Python directement dans Discord (COMMANDE MJ)

    Envoyer ``help`` dans le pseudo-terminal pour plus d'informations sur son fonctionnement.

    Évidemment, les avertissements dans ``/do`` s'appliquent ici : ne pas faire n'imp avec cette commande !!
    (même si ça peut être très utile, genre pour ajouter des gens en masse à un channel)
    """
    locs = globals()
    locs["journey"] = journey
    shell = realshell.RealShell(journey, locs)
    try:
        await shell.interact()
    except realshell.RealShellExit:
        raise commons.CommandAbortedError()


@app_commands.command()
@tools.mjs_only
@journey_command
async def co(journey: DiscordJourney, member: discord.Member):
    """Lance la procédure d'inscription pour un membre (COMMANDE MJ)

    Fat comme si on se connectait au serveur pour la première fois.

    Args:
        member: Le membre à inscrire (si son inscription a foiré).

    Cette commande est principalement destinée aux tests de développement,
    mais peut être utile si un joueur chibre son inscription
    (à utiliser dans son channel, ou ``#bienvenue`` si même le début a chibré).
    """
    await journey.send(f"Lancement du processus d'inscription pour {member.mention}", ephemeral=True)
    await features.inscription.main(member)


@app_commands.command()
@tools.mjs_only
@journey_command
async def doas(
    journey: DiscordJourney,
    *,
    joueur: app_commands.Transform[Joueur, tools.JoueurTransformer],
    command: app_commands.Transform[app_commands.Command, SubCommandTransformer],
):
    """Exécute une commande en tant qu'un joueur (COMMANDE MJ)

    Args:
        joueur: Joueur inscrit en tant que qui exécuter une commande.
        command: Commande à lancer.
    """
    if command.default_permissions:
        if command.default_permissions.manage_messages:
            await journey.send(f":x: Commande `/{command.qualified_name}` réservée aux MJs !")
            return
        if command.default_permissions.priority_speaker:
            await journey.send(f":x: Commande `/{command.qualified_name}` réservée aux MJs / rédacteurs !")
            return
        if command.default_permissions.send_messages and joueur.est_mort:
            await journey.send(f":x: Commande `/{command.qualified_name}` réservée aux joueurs en vie !")
            return

    if command.parameters:
        values = await journey.modal(
            f"/doas {joueur.nom} /{command.qualified_name}",
            *[
                discord.ui.TextInput(label=param.display_name, placeholder=param.description, required=param.required)
                for param in command.parameters
            ],
        )
    else:
        values = []

    parameters = {parameter.name: value for parameter, value in zip(command.parameters, values)}

    params_descr = " ".join(f"{name}:{value!r}" for name, value in parameters.items())
    await journey.send(
        f"/{command.qualified_name} {params_descr}", code=True, prefix=f":robot: Exécution en tant que {joueur.nom} :"
    )

    async with DiscordJourney(journey.interaction, ephemeral=True, command_author=journey.member) as journey_:
        journey_.member = joueur.member
        await command._callback._callable(journey_, **parameters)


@app_commands.command()
@tools.mjs_only
@journey_command
async def secret(journey: DiscordJourney):
    """Exécute la prochaine commande en mode "éphémère" (les messages ne s'affichent que pour le lanceur).

    Utile notamment pour faire des commandes dans un channel public, pour que la commande soit invisible.
    """
    interaction, runner = await journey.catch_next_command("Exécuter la commande à rendre silencieuse")
    async with DiscordJourney(interaction, ephemeral=True) as _journey:
        await runner(_journey)


@app_commands.command()
@journey_command
async def apropos(journey: DiscordJourney):
    """Informations et mentions légales du projet.

    N'hésitez-pas à nous contacter pour en savoir plus !
    """
    embed = discord.Embed(
        title=f"**LG-bot** - v{__version__}", description="LG-bot – Plateforme pour parties endiablées de Loup-Garou"
    )
    embed.set_author(name="À propos de ce bot :", icon_url=config.bot.user.avatar.url)
    embed.set_image(
        url=(
            "https://gist.githubusercontent.com/loic-simon/"
            "66c726053323017dba67f85d942495ef/raw/48f2607a61f3fc1b7285fd64873621035c6fbbdb/logo_espci.png"
        ),
    )
    embed.add_field(name="Auteurs", value="Loïc Simon\nTom Lacoma")
    embed.add_field(name="Licence", value="Projet open-source sous licence MIT\nhttps://opensource.org/licenses/MIT")
    embed.add_field(name="Pour en savoir plus :", value="https://github.com/GRI-ESPCI/lg-rez", inline=False)
    embed.add_field(name="Copyright :", value=":copyright: 2022 Club BD-Jeux × GRIs – ESPCI Paris - PSL", inline=False)
    embed.set_footer(text="Retrouvez-nous sur Discord : LaCarpe#1674, TaupeOrAfk#3218")

    await journey.send(embed=embed)


@app_commands.command()
@app_commands.check(lambda interaction: not config.is_setup)
@journey_command
async def setup(journey: DiscordJourney):
    """✨ Prépare un serveur nouvellement crée (COMMANDE MJ)

    À n'utiliser que dans un nouveau serveur, pour créer les rôles, catégories, salons et emojis nécessaires.
    """
    await journey.ok_cancel("Setup le serveur ?")

    original_channels = list(config.guild.channels)

    structure = config.server_structure

    # Création rôles
    await journey.send("Création des rôles...")
    roles = {}
    for slug, role in structure["roles"].items():
        roles[slug] = tools.role(role["name"], must_be_found=False)
        if roles[slug]:
            continue
        if isinstance(role["permissions"], list):
            perms = discord.Permissions(**{perm: True for perm in role["permissions"]})
        else:
            perms = getattr(discord.Permissions, role["permissions"])()
        roles[slug] = await config.guild.create_role(
            name=role["name"],
            color=int(role["color"], base=16),
            hoist=role["hoist"],
            mentionable=role["mentionable"],
            permissions=perms,
        )
    # Modification @everyone
    roles["@everyone"] = tools.role("@everyone")
    await roles["@everyone"].edit(
        permissions=discord.Permissions(**{perm: True for perm in structure["everyone_permissions"]})
    )
    await journey.send(f"{len(roles)} rôles créés.")

    # Assignation rôles
    for member in config.guild.members:
        await member.add_roles(roles["bot"] if member == config.bot.user else roles["mj"])

    # Création catégories et channels
    await journey.send("Création des salons...")
    categs = {}
    channels = {}
    for slug, categ in structure["categories"].items():
        categs[slug] = tools.channel(categ["name"], must_be_found=False)
        if not categs[slug]:
            categs[slug] = await config.guild.create_category(
                name=categ["name"],
                overwrites={
                    roles[role]: discord.PermissionOverwrite(**perms) for role, perms in categ["overwrites"].items()
                },
            )
        for position, (chan_slug, channel) in enumerate(categ["channels"].items()):
            channels[chan_slug] = tools.channel(channel["name"], must_be_found=False)
            if channels[chan_slug]:
                continue
            channels[chan_slug] = await categs[slug].create_text_channel(
                name=channel["name"],
                topic=channel["topic"],
                position=position,
                overwrites={
                    roles[role]: discord.PermissionOverwrite(**perms) for role, perms in channel["overwrites"].items()
                },
            )
        for position, (chan_slug, channel) in enumerate(categ["voice_channels"].items()):
            channels[chan_slug] = tools.channel(channel["name"], must_be_found=False)
            if channels[chan_slug]:
                continue
            channels[chan_slug] = await categs[slug].create_voice_channel(
                name=channel["name"],
                position=position,
                overwrites={roles[role]: discord.PermissionOverwrite(**perms) for role, perms in channel["overwrites"]},
            )
    await journey.send(f"{len(channels)} salons créés dans {len(categs)} catégories.")

    # Création emojis
    await journey.send("Import des emojis... (oui c'est très long)")

    async def _create_emoji(name: str, data: bytes):
        can_use = None
        if restrict := structure["emojis"]["restrict_roles"].get(name):
            can_use = [roles[role] for role in restrict]
        await config.guild.create_custom_emoji(
            name=name,
            image=data,
            roles=can_use,
        )

    n_emojis = 0
    if structure["emojis"]["drive"]:
        folder_id = structure["emojis"]["folder_path_or_id"]
        for file in gsheets.get_files_in_folder(folder_id):
            if file["extension"] != "png":
                continue
            name = file["name"].removesuffix(".png")
            if tools.emoji(name, must_be_found=False):
                continue
            data = gsheets.download_file(file["file_id"])
            await _create_emoji(name, data)
            n_emojis += 1
    else:
        root = structure["emojis"]["folder_path_or_id"]
        for file in os.scandir(root):
            name, extension = os.path.splitext(file.name)
            if extension != ".png":
                continue
            if tools.emoji(name, must_be_found=False):
                continue
            with open(file.path, "rb") as fh:
                data = fh.read()
            await _create_emoji(name, data)
            n_emojis += 1
    await journey.send(f"{n_emojis} emojis importés.")

    # Paramètres généraux du serveur
    await journey.send("Configuration du serveur...")
    if not structure["icon"]:
        icon_data = None
    elif structure["icon"]["drive"]:
        file_id = structure["icon"]["png_path_or_id"]
        icon_data = gsheets.download_file(file_id)
    else:
        with open(structure["icon"]["png_path_or_id"], "rb") as fh:
            icon_data = fh.read()

    await config.guild.edit(
        name=structure["name"],
        icon=icon_data,
        afk_channel=channels.get(structure["afk_channel"]),
        afk_timeout=int(structure["afk_timeout"]),
        verification_level=discord.VerificationLevel[structure["verification_level"]],
        default_notifications=discord.NotificationLevel[structure["default_notifications"]],
        explicit_content_filter=discord.ContentFilter[structure["explicit_content_filter"]],
        system_channel=channels[structure["system_channel"]],
        system_channel_flags=discord.SystemChannelFlags(**structure["system_channel_flags"]),
        preferred_locale=structure["preferred_locale"],
        reason="Guild set up!",
    )
    await journey.send(f"Fin de la configuration !")

    config.is_setup = True

    # Delete current chan (will also trigger on_ready)
    await journey.yes_no("Terminé ! Ce salon va être détruit (ce n'est pas une question).")
    for channel in original_channels:
        await channel.delete()


@app_commands.command()
@tools.mjs_only
@journey_command
async def command(
    journey: DiscordJourney,
    mode: Literal["enable", "disable"],
    command: app_commands.Transform[app_commands.Command | app_commands.Group, CommandTransformer],
):
    """✨ Active ou désactive une commande (COMMANDE MJ)

    Args:
        mode: Opération à réaliser.
        command: Commande à activer/désactiver (Y COMPRIS POUR LES MJS !).
    """
    if mode == "enable":
        if config.bot.tree.enable_command(command.qualified_name):
            await config.bot.tree.sync(guild=config.guild)
        await journey.send(f"Commande `/{command.qualified_name}` activée.")
    else:
        if config.bot.tree.disable_command(command.qualified_name):
            await config.bot.tree.sync(guild=config.guild)
        await journey.send(f"Commande `/{command.qualified_name}` désactivée.")



"""lg-rez / LGBot

Classe principale"""

from __future__ import annotations
import asyncio

import logging
import sys
import time
import traceback
import typing

import discord
import readycheck

from lgrez import __version__, commands, config, bdd
from lgrez.blocs import env, tools, console
from lgrez.features import gestion_ia, inscription


async def _check_and_prepare_objects(bot: LGBot) -> None:
    if not config.is_setup:
        return

    if len(config.guild.channels) < 20:
        config.is_setup = False
        # *really* needed objects, even if nothing is setup
        config.Channel.logs = config.guild.text_channels[0]
        await tools.log("Server not setup - call `/setup` !")
        config.bot.tree.enable_command("setup")
        await config.bot.tree.sync(guild=config.guild)
        return

    errors = []

    def prepare_attributes(
        rc_class: type[readycheck.ReadyCheck],
        discord_type: type[discord.abc.Snowflake],
        converter: typing.Callable[[str], typing.Any],
    ) -> None:
        """Rend prêt les attributs d'une classe ReadyCheck"""
        for attr in rc_class:
            raw = rc_class.get_raw(attr)
            # Si déjà prêt, on actualise quand même (reconnexion, update...)
            name = raw.name if isinstance(raw, discord_type) else raw
            try:
                ready = converter(name)
            except ValueError:
                qualname = f"config.{rc_class.__name__}.{attr}"
                errors.append(f"{discord_type.__name__} {qualname} = " f'"{name}" non trouvé !')
            else:
                setattr(rc_class, attr, ready)

    prepare_attributes(config.Role, discord.Role, tools.role)
    prepare_attributes(config.Channel, discord.TextChannel, tools.channel)
    prepare_attributes(config.Emoji, discord.Emoji, tools.emoji)

    try:
        tools.channel(config.private_chan_category_name)
    except ValueError:
        errors.append(
            f"catégorie config.private_chan_category_name = " f'"{config.private_chan_category_name}" non trouvée'
        )

    try:
        tools.channel(config.boudoirs_category_name)
    except ValueError:
        errors.append(f"catégorie config.boudoirs_category_name = " f'"{config.boudoirs_category_name}" non trouvée')

    try:
        tools.channel(config.old_boudoirs_category_name)
    except ValueError:
        errors.append(
            f"catégorie config.old_boudoirs_category_name = " f'"{config.old_boudoirs_category_name}" non trouvée'
        )

    if len(errors) > config._missing_objects:
        # Nouvelles erreurs
        msg = f"LGBot.on_ready: {len(errors)} errors:\n - " + "\n - ".join(errors)
        logging.error(msg)

        try:
            atmj = config.Role.mj.mention
        except readycheck.NotReadyError:
            atmj = "@everyone"

        try:
            await tools.log(msg, code=True, prefix=f"{atmj} ERREURS :")
        except readycheck.NotReadyError:
            config.Channel.logs = config.guild.text_channels[0]
            msg += "\n-- Routing logs to this channel."
            await tools.log(msg, code=True, prefix=f"{atmj} ERREURS :")

    elif len(errors) < config._missing_objects:
        if errors:
            # Erreurs résolues, il en reste
            msg = f"{len(errors)} errors:\n - " + "\n - ".join(errors)
            logging.error(msg)
            await tools.log(msg, code=True, prefix=f"Erreurs restantes :")
        else:
            # Toutes erreurs résolues
            await tools.log("Configuration rôles/chans/emojis OK.")

    config._missing_objects = len(errors)

    # Webhook
    existing = await config.Channel.logs.webhooks()

    if existing:
        config.webhook = existing[0]
    else:  # Création du webhook
        config.webhook = await config.Channel.logs.create_webhook(
            name=bot.user.name, avatar=await bot.user.avatar.read()
        )
        await tools.log(f"Webhook de tâches planifiées créé")


# ---- Définition classe principale


class LGBot(discord.Client):
    """Bot communautaire pour parties de Loup-Garou à la PCéenne.

    Classe fille de :class:`discord.ext.commands.Bot`, implémentant les
    commandes et fonctionnalités du Loup-Garou de la Rez.

    Args:
        command_prefix: passé à :class:`discord.ext.commands.Bot`
        case_insensitive: passé à :class:`discord.ext.commands.Bot`
        description: idem, défaut \: :attr:`lgrez.bot.default_description`
        \*\*kwargs: autres options de :class:`~discord.ext.commands.Bot`

    Warning:
        LG-Bot n'est **pas** thread-safe : seule une instance du bot
        peut tourner en parallèle dans un interpréteur.

        (Ceci est du aux objets de :mod:`.config`, contenant directement
        le bot, le serveur Discord, la session de connexion BDD... ;
        cette limitation résulte d'une orientation volontaire du module
        depuis sa version 2.0 pour simplifier et optimiser la
        manipulation des objects et fonctions).

    Attributes:
        GUILD_ID (int): L'ID du serveur sur lequel tourne le bot
            (normalement toujours :attr:`config.guild` ``.id``).
            Vaut ``None`` avant l'appel à :meth:`run`, puis la valeur
            de la variable d'environnement ``LGREZ_SERVER_ID``.
        in_command (list[int]): IDs des salons dans lesquels une
            commande est en cours d'exécution.
        in_stfu (list[int]): IDs des salons en mode STFU.
        in_fals (list[int]): IDs des salons en mode Foire à la saucisse.
        tasks (dict[int (.bdd.Tache.id), asyncio.TimerHandle]):
            Tâches planifiées actuellement en attente. Privilégier
            plutôt l'emploi de :attr:`.bdd.Tache.handler`.

    """

    def __init__(self, **kwargs):
        """Initialize self"""
        super().__init__(
            intents=discord.Intents.all(),
            activity=discord.Activity(type=discord.ActivityType.listening, name="vos demandes ➡️ /help"),
            **kwargs,
        )

        self.GUILD_ID = int(env.load("LGREZ_SERVER_ID"))
        self.in_stfu = []
        self.in_fals = []
        self.tasks = {}

        self.tree = commands.LGCommandTree(self, self.GUILD_ID)
        self.tree.register_commands()

    async def setup_hook(self) -> None:
        print(f"     Connected to the gateway")
        print("[3/4] Loading application (bot.setup_hook)...")

        # Déclaration des commandes
        guild = discord.Object(id=self.GUILD_ID)
        await self.tree.sync(guild=guild)
        print(f"     Application loaded: {self.application}")

    async def on_ready(self) -> None:
        """Méthode appelée par Discord au démarrage du bot.

        Vérifie le serveur (appelle :meth:`check_and_prepare_objects`),
        log et affiche publiquement que le bot est fonctionnel (activité) ;
        restaure les tâches planifiées éventuelles et exécute celles manquées.

        Si :attr:`config.output_liveness` vaut ``True``, lance le système
        :meth:`.i_am_alive` (écriture chaque minute sur un fichier disque)
        """
        if config.is_ready:
            return

        print("[4/4] Initialization (bot.on_ready)...")

        guild = self.get_guild(self.GUILD_ID)
        if not guild:
            raise RuntimeError(f"on_ready : Serveur d'ID {self.GUILD_ID} (``LGREZ_SERVER_ID``) introuvable")

        print(f"      Connected to '{guild.name}'! ({len(guild.channels)} channels, {len(guild.members)} members)")
        config.guild = guild

        # Start liveness regular output
        if config.output_liveness:
            self.i_am_alive()

        # Start admin console in the background
        asyncio.create_task(console.run_admin_console(globals()))

        # Préparations des objects globaux
        await self.check_and_prepare_objects()

        await tools.log("Just rebooted!")

        # Tâches planifiées
        taches = bdd.Tache.query.all()
        for tache in taches:
            # Si action manquée, l'exécute immédiatement, sinon l'enregistre
            tache.register()

        if taches:
            await tools.log(f"{len(taches)} tâches planifiées récupérées en base et reprogrammées.")

        config.is_ready = True
        print("      Initialization complete.")
        print("\nListening for events.")

    async def on_member_join(self, member: discord.Member) -> None:
        """Méthode appelée par Discord à l'arrivée d'un nouveau membre.

        Log et lance le processus d'inscription.

        Args:
            member: Le membre qui vient d'arriver.
        """
        await tools.log(f"Arrivée de {member.name}#{member.discriminator} sur le serveur")
        await inscription.main(member)

    async def on_member_remove(self, member: discord.Member) -> None:
        """Méthode appelée par Discord au départ d'un membre du serveur.

        Log en mentionnant les MJs.

        Args:
            member: Le joueur qui vient de partir.
        """
        await tools.log(
            f"{tools.mention_MJ(member)} ALERTE : départ du serveur de "
            f"{member.display_name} ({member.name}#{member.discriminator}) !"
        )

    async def on_message(self, message: discord.Message) -> None:
        """Méthode appelée par Discord à chaque message sur le serveur.

        Invoque l'ensemble des commandes, ou les règles d'IA si
            - Le message n'est pas une commande
            - Le message est posté dans un channel privé (dont le nom
              commence par :attr:`config.private_chan_prefix`)
            - Il n'y a pas déjà de commande en cours dans ce channel
            - Le channel n'est pas en mode STFU

        Ne fait rien si le message est envoyé par le bot lui-même
        ou par un membre sans aucun rôle affecté.

        Args:
            message: Le message qui vient d'être posté.
        """
        if not config.is_setup:
            return

        if message.author == self.user:  # Pas de boucles infinies
            return

        if not message.guild:
            await message.reply("Je n'accepte pas les messages privés, désolé !")
            return

        if message.guild != config.guild:
            return

        if not message.webhook_id and message.author.top_role == config.Role.everyone:
            # Pas de rôle affecté : le bot te calcule même pas
            return

        if message.content.startswith("/"):
            # Sécurité si fail de commande
            await message.delete()
            return

        if message.channel.name.startswith(config.private_chan_prefix) and message.channel.id not in self.in_stfu:
            # Conditions d'IA respectées (voir doc) : on trigger
            await gestion_ia.process_ia(message, message.reply)

    async def on_error(self, event: str, *args, **kwargs) -> None:
        """Méthode appelée par Discord à une exception hors commande.

        Log en mentionnant les MJs. Cette méthode permet de gérer les
        exceptions sans briser la loop du bot (i.e. il reste en ligne).

        Args:
            event: Nom de l'évènement ayant généré une erreur (``"member_join"``, ``"message"``...)
            *args, \**kwargs: Arguments passés à la fonction traitant l'évènement : ``member``, ``message``...
        """
        etype, exc, tb = sys.exc_info()  # Exception ayant causé l'appel

        if isinstance(exc, (bdd.SQLAlchemyError, bdd.DriverOperationalError)):  # Erreur SQL
            try:
                config.session.rollback()  # On rollback la session
                await tools.log("Rollback session")
            except readycheck.NotReadyError:
                pass

        await tools.log(traceback.format_exc(), code=True, prefix=f"{config.Role.mj.mention} ALED : Exception Python !")

        # On remonte l'exception à Python (pour log, ne casse pas la loop)
        raise

    async def on_guild_channel_delete(self, channel: discord.TextChannel) -> None:
        if channel.guild == config.guild:
            await self.check_and_prepare_objects()

    async def on_guild_channel_update(self, before: discord.TextChannel, after: discord.TextChannel) -> None:
        if before.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_channel_create(self, channel: discord.TextChannel) -> None:
        if channel.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_role_delete(self, role: discord.Role) -> None:
        if role.guild == config.guild:
            await self.check_and_prepare_objects()

    async def on_guild_role_update(self, before: discord.Role, after: discord.Role) -> None:
        if before.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_role_create(self, role: discord.Role) -> None:
        if role.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_emojis_update(
        self, guild: discord.Guild, before: list[discord.Emoji], after: list[discord.Emoji]
    ) -> None:
        if guild == config.guild:
            await self.check_and_prepare_objects()

    async def on_webhooks_update(self, channel: discord.TextChannel) -> None:
        if channel == config.Channel.logs:
            await self.check_and_prepare_objects()

    # Checks en temps réels des modifs des objets nécessaires au bot
    async def check_and_prepare_objects(self) -> None:
        """Vérifie et prépare les objets Discord nécessaires au bot.

        Remplit :class:`.config.Role`, :class:`.config.Channel`,
        :class:`.config.Emoji`, :attr:`config.private_chan_category_name`,
        :attr:`config.boudoirs_category_name` et :attr:`config.webhook`
        avec les objets Discord correspondants, et avertit les MJs en cas
        d'éléments manquants.
        """
        await _check_and_prepare_objects(self)

    # Système de vérification de vie
    def i_am_alive(self, filename: str = "alive.log") -> None:
        """Témoigne que le bot est en vie et non bloqué.

        Exporte le temps actuel (UTC) et planifie un nouvel appel
        dans 60s. Ce processus n'est lancé que si
        :attr:`config.output_liveness` est mis à ``True`` (*opt-in*).

        Args:
            filename: fichier où exporter le temps actuel
                (écrase le contenu).
        """
        with open(filename, "w") as f:
            f.write(str(time.time()))
        asyncio.get_running_loop().call_later(60, self.i_am_alive, filename)

    # Lancement du bot
    def run(self) -> None:
        """Prépare puis lance le bot (bloquant).

        Récupère les informations de connexion, établit la connexion
        à la base de données puis lance le bot.
        """
        print(f"--- LGBot v{__version__} ---")
        print(f"  * Front : discord.py v{discord.__version__}")

        # Récupération du token du bot et de l'ID du serveur
        LGREZ_DISCORD_TOKEN = env.load("LGREZ_DISCORD_TOKEN")

        # Connexion BDD
        print("[1/4] Connecting to database...")
        bdd.connect()
        url = config.engine.url
        print(f"      Connected to {url.host}/{url.database}!")

        # Enregistrement
        config.bot = self

        # Lancement du bot (bloquant)
        print("[2/3] Connecting to Discord...")
        super().run(LGREZ_DISCORD_TOKEN)

        print("\nDisconnected.")



"""lg-rez / bdd / Modèle de données

Déclaration de toutes les tables et leurs colonnes

"""

from __future__ import annotations

import asyncio
import datetime
import typing

import discord
import sqlalchemy
from sqlalchemy.ext.hybrid import hybrid_property

from lgrez import config
from lgrez.bdd import base, ActionTrigger
from lgrez.bdd.base import autodoc_Column, autodoc_ManyToOne, autodoc_OneToMany, autodoc_DynamicOneToMany
from lgrez.bdd.enums import UtilEtat, CibleType, Vote


class _FakeInteraction(discord.Interaction):
    __slots__ = discord.Interaction.__slots__

    class _FakeInteractionResponse:
        def is_done(self):
            return True

        async def defer(*args, **kwargs):
            pass

    class _FakeInteractionFollowup:
        def __init__(self, message: discord.Message) -> None:
            self.send = message.reply

    def __init__(self, message: discord.Message):
        self._original_response = message
        self.user = message.author
        self._cs_channel = message.channel
        self._cs_response = self._FakeInteractionResponse()
        self._cs_followup = self._FakeInteractionFollowup(message)

    @property
    def created_at(self):
        return self._original_response.created_at


class Action(base.TableBase):
    """Table de données des actions attribuées (liées à un joueur).

    Les instances doivent être enregistrées via
    :func:`.gestion_actions.add_action` et supprimées via
    :func:`.gestion_actions.delete_action`.
    """

    id: int = autodoc_Column(
        sqlalchemy.Integer(), primary_key=True, doc="Identifiant unique de l'action, sans signification"
    )

    _joueur_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("joueurs.discord_id"),
        nullable=False,
    )
    joueur: Joueur = autodoc_ManyToOne(
        "Joueur",
        back_populates="actions",
        doc="Joueur concerné",
    )

    _base_slug = sqlalchemy.Column(sqlalchemy.ForeignKey("baseactions.slug"))
    base: BaseAction | None = autodoc_ManyToOne(
        "BaseAction",
        back_populates="actions",
        nullable=True,
        doc="Action de base (``None`` si action de vote)",
    )

    vote: Vote | None = autodoc_Column(
        sqlalchemy.Enum(Vote),
        doc="Si action de vote, vote concerné",
    )

    active: bool = autodoc_Column(
        sqlalchemy.Boolean(),
        nullable=False,
        default=True,
        doc="Si l'action est actuellement utilisable (False = archives)",
    )

    cooldown: int = autodoc_Column(
        sqlalchemy.Integer(),
        nullable=False,
        default=0,
        doc="Nombre d'ouvertures avant disponibilité de l'action",
    )
    charges: int | None = autodoc_Column(
        sqlalchemy.Integer(),
        doc="Nombre de charges restantes (``None`` si illimité)",
    )

    # One-to-manys
    taches: list[Tache] = autodoc_OneToMany(
        "Tache",
        back_populates="action",
        doc="Tâches liées à cette action",
    )
    utilisations: list[Utilisation] = autodoc_DynamicOneToMany(
        "Utilisation",
        back_populates="action",
        doc="Utilisations de cette action",
    )

    def __init__(self, *args, **kwargs) -> None:
        """Initialize self."""
        n_args = ("base" in kwargs) + ("_base_slug" in kwargs) + ("vote" in kwargs)
        if not n_args:
            raise ValueError("bdd.Action: 'base'/'_base_slug' or 'vote keyword-only argument must be specified")
        elif n_args > 1:
            raise ValueError(
                "bdd.Action: 'base'/'_base_slug' and 'vote' keyword-only argument cannot both be specified"
            )
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<Action #{self.id} ({self.base or self.vote}/{self.joueur})>"

    @property
    def utilisation_ouverte(self) -> Utilisation | None:
        """Utilisation de l'action actuellement ouverte.

        Vaut ``None`` si aucune action n'a actuellement l'état
        :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.

        Raises:
            RuntimeError: plus d'une action a actuellement l'état
            :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.
        """
        filtre = Utilisation.etat.in_({UtilEtat.ouverte, UtilEtat.remplie})
        try:
            return self.utilisations.filter(filtre).one_or_none()
        except sqlalchemy.orm.exc.MultipleResultsFound:
            raise ValueError(f"Plusieurs utilisations ouvertes pour `{self}` !")

    @property
    def derniere_utilisation(self) -> Utilisation | None:
        """Dernière utilisation de cette action (temporellement).

        Considère l'utilisation ouverte le cas échéant, sinon la
        dernière utilisation par timestamp de fermeture descendant
        (quelque soit son état, y comprs :attr:`~.bdd.UtilEtat.contree`).

        Vaut ``None`` si l'action n'a jamais été utilisée.

        Raises:
            RuntimeError: plus d'une action a actuellement l'état
            :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.
        """
        return self.utilisation_ouverte or self.utilisations.order_by(Utilisation.ts_close.desc()).first()

    @property
    def decision(self) -> str:
        """Description de la décision de la dernière utilisation.

        Considère l'utilisation ouverte le cas échéant, sinon la
        dernière utilisation par timestamp de fermeture descendant.

        Vaut :attr:`.Utilisation.decision`, ou ``"<N/A>"`` si il n'y a
        aucune utilisation de cette action.

        Raises:
            RuntimeError: plus d'une action a actuellement l'état
            :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.
        """
        util = self.derniere_utilisation
        if util:
            return util.decision
        else:
            return "<N/A>"

    @hybrid_property
    def is_open(self) -> bool:
        """L'action est ouverte (l'utilisateur peut interagir) ?

        *I.e.* l'action a au moins une utilisation
        :attr:`~.bdd.UtilEtat.ouverte` ou :attr:`~.bdd.UtilEtat.remplie`.

        Propriété hybride (:class:`sqlalchemy.ext.hybrid.hybrid_property`) :

            - Sur l'instance, renvoie directement la valeur booléenne ;
            - Sur la classe, renvoie la clause permettant de déterminer
              si l'action est en attente.

        Examples::

            action.is_open          # bool
            Joueur.query.filter(Joueur.actions.any(Action.is_open)).all()
        """
        return bool(self.utilisations.filter(Utilisation.is_open).all())

    @is_open.expression
    def is_open(cls) -> sqlalchemy.sql.selectable.Exists:
        return cls.utilisations.any(Utilisation.is_open)

    @hybrid_property
    def is_waiting(self) -> bool:
        """L'action est ouverte et aucune décision n'a été prise ?

        *I.e.* la clause a au moins une utilisation
        :attr:`~.bdd.UtilEtat.ouverte`.

        Propriété hybride (voir :attr:`.is_open` pour plus d'infos)
        """
        return bool(self.utilisations.filter(Utilisation.is_waiting).all())

    @is_waiting.expression
    def is_waiting(cls) -> sqlalchemy.sql.selectable.Exists:
        return cls.utilisations.any(Utilisation.is_waiting)


class Utilisation(base.TableBase):
    """Table de données des utilisations des actions.

    Les instances sont enregistrées via :meth:`\/open
    <.open_close.OpenClose.OpenClose.open.callback>` ;
    elles n'ont pas vocation à être supprimées.
    """

    id: int = autodoc_Column(
        sqlalchemy.BigInteger(),
        primary_key=True,
        doc="Identifiant unique de l'utilisation, sans signification",
    )

    _action_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("actions.id"),
        nullable=False,
    )
    action: Action = autodoc_ManyToOne(
        "Action",
        back_populates="utilisations",
        doc="Action utilisée",
    )

    etat: UtilEtat = autodoc_Column(
        sqlalchemy.Enum(UtilEtat),
        nullable=False,
        default=UtilEtat.ouverte,
        doc="État de l'utilisation",
    )

    ts_open: datetime.datetime | None = autodoc_Column(
        sqlalchemy.DateTime(),
        doc="Timestamp d'ouverture de l'utilisation",
    )
    ts_close: datetime.datetime | None = autodoc_Column(
        sqlalchemy.DateTime(),
        doc="Timestamp de fermeture de l'utilisation",
    )
    ts_decision: datetime.datetime | None = autodoc_Column(
        sqlalchemy.DateTime(),
        doc="Timestamp du dernier remplissage de l'utilisation",
    )

    # One-to-manys
    ciblages: list[Ciblage] = autodoc_OneToMany(
        "Ciblage",
        back_populates="utilisation",
        doc="Cibles désignées dans cette utilisation",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<Utilisation #{self.id} ({self.action}/{self.etat})>"

    def open(self) -> None:
        """Ouvre cette utilisation.

        Modifie son :attr:`etat`, définit :attr:`ts_open` au temps
        actuel, et update.
        """
        self.etat = UtilEtat.ouverte
        self.ts_open = datetime.datetime.now()
        self.update()

    def close(self) -> None:
        """Clôture cette utilisation.

        Modifie son :attr:`etat`, définit :attr:`ts_close` au temps
        actuel, et update.
        """
        if self.etat == UtilEtat.remplie:
            self.etat = UtilEtat.validee
        else:
            self.etat = UtilEtat.ignoree
        self.ts_close = datetime.datetime.now()
        self.update()

    def ciblage(self, slug: str) -> Ciblage:
        """Renvoie le ciblage de cette utilisation de slug voulu.

        Args:
            slug: Doit correspondre à un des slugs des bases
                des :attr:`ciblages` de l'utilisation.

        Returns:
            Le ciblage de slug correspondant.

        Raises:
            ValueError: slug non trouvé dans les :attr:`ciblages`.
        """
        try:
            return next(cib for cib in self.ciblages if cib.base.slug == slug)
        except StopIteration:
            raise ValueError(f"{self} : pas de ciblage de slug '{slug}'") from None

    @property
    def cible(self) -> Joueur | None:
        """Joueur ciblé par l'utilisation, si applicable.

        Cet attribut n'est accessible que si l'utilisation est d'un vote
        ou d'une action définissant un et une seul ciblage de type
        :attr:`~bdd.CibleType.joueur`, :attr:`~bdd.CibleType.vivant`
        ou :attr:`~bdd.CibleType.mort`.

        Vaut ``None`` si l'utilisation a l'état
        :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.ignoree`.

        Raises:
            ValueError: l'action ne remplit pas les critères évoqués
            ci-dessus
        """
        if self.action.vote:
            # vote : un BaseCiblage implicite de type CibleType.vivants
            return self.ciblages[0].joueur if self.ciblages else None
        else:
            base_ciblages = self.action.base.base_ciblages
            bc_joueurs = [bc for bc in base_ciblages if bc.type in [CibleType.joueur, CibleType.vivant, CibleType.mort]]
            if len(bc_joueurs) != 1:
                raise ValueError(f"L'utilisation {self} n'a pas une et une seule cible de type joueur")

            base_ciblage = bc_joueurs[0]
            try:
                ciblage = next(cib for cib in self.ciblages if cib.base == base_ciblage)
            except StopIteration:
                return None  # Pas de ciblage fait

            return ciblage.joueur

    @property
    def decision(self) -> str:
        """Description de la décision de cette utilisation.

        Complète le template de :.bdd.BaseAction.decision_format` avec
        les valeurs des ciblages de l'utilisation.

        Vaut ``"Ne rien faire"`` si l'utilisation n'a pas de ciblages,
        et :attr:`.cible` dans le cas d'un vote.
        """
        if not self.action.base:
            return str(self.cible)

        if not self.ciblages:
            return "Ne rien faire"

        template = self.action.base.decision_format
        data = {ciblage.base.slug: ciblage.valeur_descr for ciblage in self.ciblages}
        try:
            return template.format(**data)
        except KeyError:
            return template

    @hybrid_property
    def is_open(self) -> bool:
        """L'utilisation est ouverte (l'utilisateur peut interagir) ?

        Raccourci pour
        ``utilisation.etat in {UtilEtat.ouverte, UtilEtat.remplie}``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return self.etat in {UtilEtat.ouverte, UtilEtat.remplie}

    @is_open.expression
    def is_open(cls) -> sqlalchemy.sql.selectable.Exists:
        return cls.etat.in_({UtilEtat.ouverte, UtilEtat.remplie})

    @hybrid_property
    def is_waiting(self) -> bool:
        """L'utilisation est ouverte et aucune décision n'a été prise ?

        Raccourci pour ``utilisation.etat == UtilEtat.ouverte``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return self.etat == UtilEtat.ouverte

    @is_waiting.expression
    def is_waiting(cls) -> sqlalchemy.sql.selectable.Exists:
        return cls.etat == UtilEtat.ouverte

    @hybrid_property
    def is_filled(self) -> bool:
        """L'utilisation est remplie (l'utilisateur a interagi avec) ?

        Raccourci pour
        ``utilisation.etat in {UtilEtat.remplie, UtilEtat.validee,
        UtilEtat.contree}``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return self.etat in {UtilEtat.remplie, UtilEtat.validee, UtilEtat.contree}

    @is_filled.expression
    def is_filled(cls) -> sqlalchemy.sql.selectable.Exists:
        return cls.etat.in_({UtilEtat.remplie, UtilEtat.validee, UtilEtat.contree})


class Ciblage(base.TableBase):
    """Table de données des cibles désignées dans les utilisations d'actions.

    Les instances sont enregistrées via :meth:`\/action
    <.voter_agir.VoterAgir.VoterAgir.action.callback>` ;
    elles n'ont pas vocation à être supprimées.
    """

    id: int = autodoc_Column(
        sqlalchemy.Integer(),
        primary_key=True,
        doc="Identifiant unique du ciblage, sans signification",
    )

    _base_id = sqlalchemy.Column(sqlalchemy.ForeignKey("baseciblages.id"))
    base: BaseCiblage | None = autodoc_ManyToOne(
        "BaseCiblage",
        back_populates="ciblages",
        nullable=True,
        doc="Modèle de ciblage (lié au modèle d'action). Vaut ``None`` pour un ciblage de vote",
    )

    _utilisation_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("utilisations.id"),
        nullable=False,
    )
    utilisation: Utilisation = autodoc_ManyToOne(
        "Utilisation",
        back_populates="ciblages",
        doc="Utilisation où ce ciblage a été fait",
    )

    _joueur_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("joueurs.discord_id"),
        nullable=True,
    )
    joueur: Joueur | None = autodoc_ManyToOne(
        "Joueur",
        back_populates="ciblages",
        nullable=True,
        doc="Joueur désigné, si ``base.type`` vaut "
        ":attr:`~.bdd.CibleType.joueur`, :attr:`~.bdd.CibleType.vivant` "
        "ou :attr:`~.bdd.CibleType.mort`",
    )

    _role_slug = sqlalchemy.Column(
        sqlalchemy.ForeignKey("roles.slug"),
        nullable=True,
    )
    role: Role | None = autodoc_ManyToOne(
        "Role",
        back_populates="ciblages",
        nullable=True,
        doc="Rôle désigné, si ``base.type`` vaut :attr:`~.bdd.CibleType.role`",
    )

    _camp_slug = sqlalchemy.Column(
        sqlalchemy.ForeignKey("camps.slug"),
        nullable=True,
    )
    camp: Camp | None = autodoc_ManyToOne(
        "Camp",
        back_populates="ciblages",
        nullable=True,
        doc="Camp désigné, si ``base.type`` vaut :attr:`~.bdd.CibleType.camp`",
    )

    booleen: bool | None = autodoc_Column(
        sqlalchemy.Boolean(),
        doc="Valeur, si ``base.type`` vaut :attr:`~.bdd.CibleType.booleen`",
    )
    texte: str | None = autodoc_Column(
        sqlalchemy.String(1000),
        doc="Valeur, si ``base.type`` vaut :attr:`~.bdd.CibleType.texte`",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<Ciblage #{self.id} ({self.base}/{self.utilisation})>"

    @property
    def _val_attr(self) -> str:
        """Nom de l'attribut stockant la valeur du ciblage"""
        if not self.base or self.base.type in {CibleType.joueur, CibleType.vivant, CibleType.mort}:  # vote
            return "joueur"
        elif self.base.type == CibleType.role:
            return "role"
        elif self.base.type == CibleType.camp:
            return "camp"
        elif self.base.type == CibleType.booleen:
            return "booleen"
        elif self.base.type == CibleType.texte:
            return "texte"
        else:
            raise ValueError(f"Ciblage de type inconnu : {self.base.type}")

    @property
    def valeur(self) -> Joueur | Role | Camp | bool | str:
        """Valeur du ciblage, selon son type.

        Propriété en lecture et écriture.

        Raises:
            ValueError: ciblage de type inconnu
        """
        return getattr(self, self._val_attr)

    @valeur.setter
    def valeur(self, value):
        setattr(self, self._val_attr, value)

    @property
    def valeur_descr(self) -> str:
        """Description de la valeur du ciblage.

        Si :attr:`valeur` vaut ``None``, renvoie ``<N/A>``

        Raises:
            ValueError: ciblage de type inconnu
        """
        if self.valeur is None:
            return "<N/A>"

        if not self.base or self.base.type in {CibleType.joueur, CibleType.vivant, CibleType.mort}:  # vote
            return self.joueur.nom
        elif self.base.type == CibleType.role:
            return self.role.nom_complet
        elif self.base.type == CibleType.camp:
            return self.camp.nom
        elif self.base.type == CibleType.booleen:
            return "Oui" if self.booleen else "Non"
        else:
            return self.texte


class Tache(base.TableBase):
    """Table de données des tâches planifiées du bot.

    Les instances doivent être enregistrées via :meth:`.add`
    et supprimées via :func:`.delete`.
    """

    _tasks = set()

    id: int = autodoc_Column(
        sqlalchemy.Integer(),
        primary_key=True,
        doc="Identifiant unique de la tâche, sans signification",
    )
    timestamp: datetime.datetime = autodoc_Column(
        sqlalchemy.DateTime(),
        nullable=False,
        doc="Moment où exécuter la tâche",
    )
    commande: str = autodoc_Column(
        sqlalchemy.String(2000),
        nullable=False,
        doc="Nom complet ('command subcommand', sans le /) de la commande à exécuter",
    )
    parameters: dict | None = autodoc_Column(
        sqlalchemy.JSON(),
        nullable=True,
        doc="Paramètres de la commande à exécuter",
    )

    _action_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("actions.id"),
        nullable=True,
    )
    action: Action | None = autodoc_ManyToOne(
        "Action",
        back_populates="taches",
        nullable=True,
        doc="Si la tâche est liée à une action, action concernée",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<Tache #{self.id} ({self.description})>"

    @property
    def description(self) -> str:
        command_descr = f"/{self.commande}"
        if self.parameters:
            command_descr += " " + " ".join(str(param) for param in self.parameters.values())
        return command_descr

    @property
    def handler(self) -> asyncio.TimerHandle:
        """Représentation dans le bot de la tâche.

        Proxy pour :attr:`config.bot.tasks[self.id] <.LGBot.tasks>`,
        en lecture, écriture et suppression (``del``).

        Raises:
            RuntimeError: tâche non enregistrée dans le bot.
        """
        try:
            return config.bot.tasks[self.id]
        except KeyError:
            raise RuntimeError(f"Tâche {self} non enregistrée dans le bot !")

    @handler.setter
    def handler(self, value):
        if self.id is None:
            raise RuntimeError("Tache.handler: Tache.id non défini (commit ?)")
        config.bot.tasks[self.id] = value

    @handler.deleter
    def handler(self):
        try:
            del config.bot.tasks[self.id]
        except KeyError:
            pass

    async def invoke_command(self, tries: int = 0) -> None:
        """Exécute la tâche (coroutine programmée par :meth:`execute`).

        Envoie un webhook (:obj:`.config.webhook`) de contenu
        :attr:`commande`.

        Si une exception quelconque est levée par l'envoi du webhook,
        re-programme l'exécution de la tâche (:meth:`execute`) 2 secondes
        après ; ce jusqu'à 5 fois, après quoi un message d'alerte est
        envoyé dans :attr:`.config.Channel.logs`.

        Si aucune exception n'est levée (succès), supprime la tâche.

        Args:
            tries: Numéro de l'essai d'envoi actuellement en cours.
        """
        try:
            message = await config.Channel.logs.send(f"⏰ `#{self.id}` :arrow_forward: `{self.description}`")
            command = config.bot.tree.get_command_by_name(self.commande)
            if not command:
                await message.reply(
                    f"{config.Role.mj.mention} ALERT: commande inconnue ou désactivée : `{self.commande}`"
                )
                return
            interaction = _FakeInteraction(message)
            await command.callback(interaction, **self.parameters or {})

        except Exception as exc:
            if tries < 5:
                # On réessaie
                asyncio.get_running_loop().call_later(2, self.execute, tries + 1)
            else:
                await config.Channel.logs.send(
                    f"{config.Role.mj.mention} ALERT: impossible d'exécuter la tâche programmée "
                    f"(5 essais, erreur : ```{type(exc).__name__}: {exc})```\n"
                    f"Commande non envoyée : `{self.description}`"
                )
        else:
            self.delete()

    def execute(self, tries: int = 0) -> None:
        """Exécute la tâche planifiée (méthode appelée par la loop).

        Programme :meth:`invoke_command` pour exécution immédiate.

        Args:
            tries: Numéro de l'essai d'envoi actuellement en cours,
                passé à :meth:`invoke_command`.
        """
        task = asyncio.create_task(self.invoke_command(tries=tries))  # Programme la coroutine pour exécution immédiate
        self._tasks.add(task)  # Cf. https://docs.python.org/fr/3/library/asyncio-task.html#asyncio.create_task
        task.add_done_callback(self._tasks.discard)

    def register(self) -> None:
        """Programme l'exécution de la tâche dans la boucle d'événements du bot."""
        now = datetime.datetime.now()
        delay = (self.timestamp - now).total_seconds()
        handler = asyncio.get_running_loop().call_later(delay, self.execute)
        self.handler = handler  # TimerHandle, pour pouvoir cancel

    def cancel(self) -> None:
        """Annule et nettoie la tâche planifiée (sans la supprimer en base).

        Si la tâche a déjà été exécutée, ne fait que nettoyer le handler.
        """
        try:
            self.handler.cancel()  # Annule la tâche, pas d'effet si elle a déjà été exécutée
        except RuntimeError:  # Tâche non enregistrée
            pass
        else:
            del self.handler

    _T = typing.TypeVar("_T", bound="Tache")

    def add(self: _T, *other: _T) -> None:
        """Enregistre la tâche sur le bot et en base.

        Globalement équivalent à un appel à :meth:`.register` (pour
        chaque élément le cas échéant) avant l'ajout en base habituel
        (:meth:`TableBase.add <.bdd.base.TableBase.add>`).

        Args:
            \*other: autres instances à ajouter dans le même commit,
                éventuellement.
        """
        super().add(*other)  # Enregistre tout en base

        self.register()  # Enregistre sur le bot
        for item in other:  # Les autres aussi
            item.register()

    def delete(self: _T, *other: _T) -> None:
        """Annule la tâche planifiée et la supprime en base.

        Globalement équivalent à un appel à :meth:`.cancel` (pour
        chaque élément le cas échéant) avant la suppression en base
        habituelle (:meth:`TableBase.add <.bdd.base.TableBase.add>`).

        Args:
            \*other: autres instances à supprimer dans le même commit,
                éventuellement.
        """
        self.cancel()  # Annule la tâche
        for item in other:  # Les autres aussi
            item.cancel()

        super().delete(*other)  # Supprime tout en base


from lgrez.bdd import Joueur, Role, Camp, BaseAction, BaseCiblage



"""lg-rez / bdd / Modèle de données

Déclaration de toutes les tables et leurs colonnes

"""

from __future__ import annotations

import datetime
import typing

import discord
import sqlalchemy
from sqlalchemy.ext.hybrid import hybrid_property

from lgrez import config
from lgrez.bdd import base
from lgrez.bdd.base import autodoc_Column, autodoc_ManyToOne, autodoc_OneToMany, autodoc_DynamicOneToMany
from lgrez.bdd.enums import Statut, CandidHaroType, Vote

if typing.TYPE_CHECKING:
    from lgrez.bdd import Role, Camp, Action, Ciblage


# Tables de données


class Joueur(base.TableBase):
    """Table de données des joueurs inscrits.

    Les instances de cette classe correspondent aux lignes du Tableau de
    bord ; elles sont crées par l'inscription (:func:`.inscription.main`)
    et synchronisées par :meth:`\/sync <.sync.Sync.Sync.sync.callback>`.
    """

    discord_id: int = autodoc_Column(
        sqlalchemy.BigInteger(),
        primary_key=True,
        autoincrement=False,
        doc="ID Discord du joueur",
    )
    chan_id_: int = autodoc_Column(
        sqlalchemy.BigInteger(),
        nullable=False,
        doc="ID du chan privé Discord du joueur. *(le ``_`` final indique ,"
        "que ce champ n'est pas synchronisé avec le Tableau de bord)*",
    )

    nom: str = autodoc_Column(
        sqlalchemy.String(32),
        nullable=False,
        doc="Nom du joueur (demandé à l'inscription)",
    )
    chambre: str | None = autodoc_Column(
        sqlalchemy.String(200),
        doc="Emplacement du joueur (demandé à l'inscription)",
    )
    statut: Statut = autodoc_Column(
        sqlalchemy.Enum(Statut),
        nullable=False,
        default=Statut.vivant,
        doc="Statut RP",
    )

    _role_slug = sqlalchemy.Column(
        sqlalchemy.ForeignKey("roles.slug"),
        nullable=False,
        default=lambda: config.default_role_slug,
    )
    role: Role = autodoc_ManyToOne(
        "Role",
        back_populates="joueurs",
        doc="Rôle du joueur (défaut :meth:`.bdd.Role.default`)",
    )

    _camp_slug = sqlalchemy.Column(
        sqlalchemy.ForeignKey("camps.slug"),
        nullable=False,
        default=lambda: config.default_camp_slug,
    )
    camp: Camp = autodoc_ManyToOne(
        "Camp",
        back_populates="joueurs",
        doc="Camp du joueur (défaut :meth:`.bdd.Camp.default`)",
    )

    votant_village: bool = autodoc_Column(
        sqlalchemy.Boolean(),
        nullable=False,
        default=True,
        doc="Le joueur participe aux votes du village ?",
    )
    votant_loups: bool = autodoc_Column(
        sqlalchemy.Boolean(),
        nullable=False,
        default=False,
        doc="Le joueur participe au vote des loups ?",
    )
    role_actif: bool = autodoc_Column(
        sqlalchemy.Boolean(),
        nullable=False,
        default=True,
        doc="Le joueur peut agir ? (pas chatgarouté...)",
    )

    # One-to-manys
    actions: list[Action] = autodoc_OneToMany(
        "Action",
        back_populates="joueur",
        doc="Actions pour ce joueur",
    )
    candidharos: list[CandidHaro] = autodoc_OneToMany(
        "CandidHaro",
        back_populates="joueur",
        doc="Candidatures et haros de/contre ce joueur",
    )
    bouderies: list[Bouderie] = autodoc_OneToMany(
        "Bouderie",
        back_populates="joueur",
        doc="Appartenances aux boudoirs de ce joueur",
    )
    ciblages: list[Ciblage] = autodoc_DynamicOneToMany(
        "Ciblage",
        back_populates="joueur",
        doc="Ciblages prenant ce joueur pour cible",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<Joueur #{self.discord_id} ({self.nom})>"

    def __str__(self) -> str:
        """Return str(self)."""
        return str(self.nom)

    @property
    def member(self) -> discord.Member:
        """Membre Discord correspondant à ce Joueur.

        Raises:
            ValueError: pas de membre correspondant
            ~readycheck.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        result = config.guild.get_member(self.discord_id)
        if not result:
            raise ValueError(f"Joueur.member : pas de membre pour `{self}` !")

        return result

    @property
    def private_chan(self) -> discord.TextChannel:
        """Channel privé (conversation bot) du joueur.

        Raises:
            ValueError: pas de channel correspondant
            ~readycheck.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        result = config.guild.get_channel(self.chan_id_)
        if not result:
            raise ValueError(f"Joueur.private_chan : pas de chan pour `{self}` !")

        return result

    @property
    def actions_actives(self) -> list[Action]:
        """Sous-ensemble de :attr:`actions` restreint aux actions actives.

        Élimine aussi les actions de vote (sans base).
        """
        return [ac for ac in self.actions if ac.active and not ac.vote]

    @property
    def boudoirs(self) -> list[Boudoir]:
        """Boudoirs où est ce jour (read-only)"""
        return [boud.boudoir for boud in self.bouderies]

    @hybrid_property
    def est_vivant(self) -> bool:
        """Le joueur est en vie ou MVtisé ?

        Raccourci pour
        ``joueur.statut in {Statut.vivant, Statut.MV}``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return self.statut in {Statut.vivant, Statut.MV}

    @est_vivant.expression
    def est_vivant(cls) -> sqlalchemy.sql.selectable.Exists:
        return cls.statut.in_({Statut.vivant, Statut.MV})

    @hybrid_property
    def est_mort(self) -> bool:
        """Le joueur est mort ?

        Raccourci pour ``joueur.statut == Statut.mort``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return self.statut == Statut.mort

    @est_mort.expression
    def est_mort(cls) -> sqlalchemy.sql.selectable.Exists:
        return cls.statut == Statut.mort

    @classmethod
    def from_member(cls, member: discord.Member) -> Joueur:
        """Récupère le Joueur (instance de BDD) lié à un membre Discord.

        Args:
            member: le membre concerné.

        Returns:
            Le joueur correspondant.

        Raises:
            ValueError: membre introuvable en base
            ~readycheck.NotReadyError: session non initialisée
                (:obj:`.config.session` vaut ``None``)
        """
        joueur = cls.query.get(member.id)
        if not joueur:
            raise ValueError(f"Joueur.from_member : pas de joueur en base pour `{member}` !")

        return joueur

    def action_vote(self, vote: Vote) -> Action:
        """Retourne l'"action de vote" voulue pour ce joueur.

        Args:
            vote: vote pour lequel récupérer l'action.

        Returns:
            L'action permettant au joueur de voter.

        Raises:
            RuntimeError: action non existante
        """
        if not isinstance(vote, Vote):
            vote = Vote[vote]
        try:
            return next(act for act in self.actions if act.vote == vote)
        except StopIteration:
            raise RuntimeError(f"{self} : pas d'action de vote {vote.name} ! ``/cparti`` a bien été appelé ?") from None


class CandidHaro(base.TableBase):
    """Table de données des candidatures et haros en cours #PhilippeCandidHaro.

    Les instances sont enregistrées via :meth:`\/haro
    <.actions_publiques.ActionsPubliques.ActionsPubliques.haro.callback>`
    / :meth:`\/candid
    <.actions_publiques.ActionsPubliques.ActionsPubliques.candid.callback>`
    et supprimées via :meth:`\/wipe
    <.actions_publiques.ActionsPubliques.ActionsPubliques.wipe.callback>`.
    """

    id: int = autodoc_Column(
        sqlalchemy.Integer(),
        primary_key=True,
        doc="Identifiant unique du candidharo, sans signification",
    )

    _joueur_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("joueurs.discord_id"),
        nullable=False,
    )
    joueur: Joueur = autodoc_ManyToOne(
        "Joueur",
        back_populates="candidharos",
        doc="Joueur concerné (candidat ou haroté)",
    )

    type: CandidHaroType = autodoc_Column(
        sqlalchemy.Enum(CandidHaroType),
        nullable=False,
        doc="Haro ou candidature ?",
    )

    message_id: int | None = autodoc_Column(
        sqlalchemy.BigInteger(),
        nullable=True,
        doc="ID du message de haro associé",
    )

    async def fetch_message(self) -> discord.Message | None:
        """Le message associé à ce haro"""
        if not self.message_id:
            return None

        try:
            return await config.Channel.haros.fetch_message(self.message_id)
        except discord.NotFound:
            return None

    async def disable_message_buttons(self) -> None:
        if message := await self.fetch_message():
            view = discord.ui.View.from_message(message)
            for component in view.children:
                if isinstance(component, (discord.ui.Button | discord.ui.Select)):
                    component.disabled = True
                else:
                    view.remove_item(component)

            await message.edit(view=view)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<CandidHaro #{self.id} ({self.joueur}/{self.type})>"


class Boudoir(base.TableBase):
    """Table de données des boudoirs sur le serveur.

    Les instances de cette classe sont crées, modifiées et supprimées par
    :meth:`\/boudoir <.chans.GestionChans.GestionChans.boudoir.callback>`.
    """

    chan_id: int = autodoc_Column(
        sqlalchemy.BigInteger(),
        primary_key=True,
        autoincrement=False,
        doc="ID Discord du salon",
    )

    nom: str = autodoc_Column(
        sqlalchemy.String(80),
        nullable=False,
        doc="Nom du boudoir (demandé à la création)",
    )
    ts_created: datetime.datetime = autodoc_Column(
        sqlalchemy.DateTime(),
        nullable=False,
        doc="Timestamp de la création",
    )

    # One-to-manys
    bouderies: list[Bouderie] = autodoc_OneToMany(
        "Bouderie",
        back_populates="boudoir",
        doc="Appartenances à ce boudoir",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<Boudoir #{self.chan_id} ({self.nom})>"

    @property
    def chan(self) -> discord.TextChannel:
        """Salon Discord correspondant à ce boudoir.

        Raises:
            ValueError: pas de membre correspondant
            ~readycheck.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        result = config.guild.get_channel(self.chan_id)
        if not result:
            raise ValueError(f"Boudoir.chan : pas de chan pour {self} !")

        return result

    @property
    def joueurs(self) -> list[Joueur]:
        """Joueurs présents dans ce boudoir (read-only)"""
        return [bouderie.joueur for bouderie in self.bouderies]

    @property
    def gerant(self) -> Joueur:
        """Membre du boudoir ayant les droits de gestion (lecture & écriture).

        Raises:
            ValueError: pas de membre avec les droits de gestion
        """
        try:
            return next(bouderie.joueur for bouderie in self.bouderies if bouderie.gerant)
        except StopIteration:
            raise ValueError(f"Pas de membre gérant le boudoir *{self.nom}*")

    @gerant.setter
    def gerant(self, new_gerant: Joueur) -> None:
        bouderie_new = next(
            (boud for boud in self.bouderies if boud.joueur == new_gerant),
            None,
        )
        if not bouderie_new:
            raise ValueError(f"{new_gerant} is not in {self}")
        bouderie_old = next(boud for boud in self.bouderies if boud.gerant)
        if bouderie_old.joueur == new_gerant:
            return
        bouderie_old.gerant = False
        bouderie_new.gerant = True
        bouderie_new.ts_promu = datetime.datetime.now()

    @classmethod
    def from_channel(cls, channel: discord.TextChannel) -> Boudoir:
        """Récupère le Boudoir (instance de BDD) lié à un salon Discord.

        Args:
            channel: le salon concerné.

        Returns:
            Le salon correspondant.

        Raises:
            ValueError: boudoir introuvable en base
            ~readycheck.NotReadyError: session non initialisée
                (:obj:`.config.session` vaut ``None``)
        """
        boudoir = cls.query.get(channel.id)
        if not boudoir:
            raise ValueError(f"Boudoir.from_channel : pas de boudoir en base pour `{channel.mention}` !")

        return boudoir


class Bouderie(base.TableBase):
    """Table de données des appartenances aux boudoirs du serveur.

    Table d'association entre :class:`.Joueur` et :class:`.Boudoir`.

    Les instances de cette classe sont crées, modifiées et supprimées par
    :meth:`\/boudoir <.chans.GestionChans.GestionChans.boudoir.callback>`.
    """

    id: int = autodoc_Column(
        sqlalchemy.Integer(),
        primary_key=True,
        doc="Identifiant unique de la bouderie, sans signification",
    )

    _joueur_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("joueurs.discord_id"),
        nullable=False,
    )
    joueur: Joueur = autodoc_ManyToOne(
        "Joueur",
        back_populates="bouderies",
        doc="Joueur concerné",
    )

    _boudoir_id = sqlalchemy.Column(
        sqlalchemy.ForeignKey("boudoirs.chan_id"),
        nullable=False,
    )
    boudoir: Boudoir = autodoc_ManyToOne(
        "Boudoir",
        back_populates="bouderies",
        doc="Boudoir concerné",
    )

    gerant: bool = autodoc_Column(
        sqlalchemy.Boolean(),
        nullable=False,
        default=False,
        doc="Si le joueur a les droits de gestion du boudoir",
    )

    ts_added: datetime.datetime = autodoc_Column(
        sqlalchemy.DateTime(),
        nullable=False,
        doc="Timestamp de l'ajout du joueur au boudoir",
    )
    ts_promu: datetime.datetime | None = autodoc_Column(
        sqlalchemy.DateTime(),
        doc="Timestamp de la promotion en gérant, le cas échéant",
    )

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"<Bouderie #{self.id} ({self.joueur}/{self.boudoir})>"



"""lg-rez / features / Commandes de gestion des salons

Création, ajout, suppression de membres

"""

import asyncio
import functools
import datetime
from typing import Callable, Literal

import discord
from discord import app_commands

from lgrez import config
from lgrez.blocs import tools
from lgrez.blocs.journey import DiscordJourney, journey_command, journey_context_menu
from lgrez.bdd import Joueur, Boudoir, Bouderie
from lgrez.features.sync import transtype


def in_boudoir(callback: Callable) -> Callable:
    """Décorateur : commande utilisable dans un boudoir uniquement.

    Lors d'une invocation de la commande décorée hors d'un boudoir
    (enregistré dans :class:`.bdd.Boudoir`), affiche un message d'erreur.

    Ce décorateur n'est utilisable que sur une commande définie dans un Cog.
    """

    @functools.wraps(callback)
    async def new_callback(journey: DiscordJourney, **kwargs):
        try:
            Boudoir.from_channel(journey.channel)
        except ValueError:
            await journey.send(":x: Cette commande est invalide en dehors d'un boudoir.", ephemeral=True)
        else:
            await callback(journey, **kwargs)

    return new_callback


def gerant_only(callback: Callable) -> Callable:
    """Décorateur : commande utilisable par le gérant d'un boudoir uniquement.

    Lors d'une invocation de la commande décorée par un membre qui n'est
    pas gérant du boudoir, affiche un message d'erreur.

    Ce décorateur doit toujours être utilisé en combinaison avec
    :func:`in_boudoir` et positionné après lui.

    Ce décorateur n'est utilisable que sur une commande définie dans un Cog.
    """

    @functools.wraps(callback)
    async def new_callback(journey: DiscordJourney, **kwargs):
        boudoir = Boudoir.from_channel(journey.channel)
        gerant = Joueur.from_member(journey.member)
        if boudoir.gerant != gerant:
            await journey.send(":x: Seul le gérant du boudoir peut utiliser cette commande.", ephemeral=True)
        else:
            await callback(journey, **kwargs)

    return new_callback


async def add_joueur_to_boudoir(boudoir: Boudoir, joueur: Joueur, gerant: bool = False) -> bool:
    """Ajoute un joueur sur un boudoir.

    Crée la :class:`.Bouderie` correspondante et modifie les permissions du salon.

    Args:
        boudoir: Le boudoir où ajouter un joueur.
        joueur: Le joueur à ajouter.
        gerant: Si le joueur doit être ajouté avec les permissions de gérant.

    Returns:
        ``True`` si le joueur a été ajouté, ``False`` si il y était déjà / le boudoir est fermé.
    """
    if joueur in boudoir.joueurs:
        # Joueur déjà dans le boudoir
        return False
    if not boudoir.joueurs and not gerant:
        # Boudoir fermé (plus de joueurs) et pas ajout comme gérant
        return False

    now = datetime.datetime.now()
    Bouderie(boudoir=boudoir, joueur=joueur, gerant=gerant, ts_added=now, ts_promu=now if gerant else None).add()
    await boudoir.chan.set_permissions(joueur.member, read_messages=True)

    # Sortie du cimetière le cas échéant
    if tools.in_multicateg(boudoir.chan.category, config.old_boudoirs_category_name):
        await boudoir.chan.send(tools.ital("[Ce boudoir contient au moins deux joueurs vivants, " "désarchivage...]"))
        categ = await tools.multicateg(config.boudoirs_category_name)
        await boudoir.chan.edit(name=boudoir.nom, category=categ)
    return True


async def remove_joueur_from_boudoir(boudoir: Boudoir, joueur: Joueur) -> None:
    """Retire un joueur d'un boudoir.

    Supprime la :class:`.Bouderie` correspondante et modifie les permissions du salon.

    Args:
        boudoir: Le boudoir d'où enlever un joueur.
        joueur: Le joueur à enlever.
    """
    Bouderie.query.filter_by(boudoir=boudoir, joueur=joueur).one().delete()
    await boudoir.chan.set_permissions(joueur.member, overwrite=None)
    # Déplacement dans le cimetière si nécessaire
    vivants = [jr for jr in boudoir.joueurs if jr.est_vivant]
    if len(vivants) < 2:
        if tools.in_multicateg(boudoir.chan.category, config.old_boudoirs_category_name):
            # Boudoir déjà au cimetière
            return
        await boudoir.chan.send(tools.ital("[Ce boudoir contient moins de deux joueurs vivants, " "archivage...]"))
        categ = await tools.multicateg(config.old_boudoirs_category_name)
        await boudoir.chan.edit(
            name=f"\N{CROSS MARK} {boudoir.nom}",
            category=categ,
        )


async def _create_boudoir(gerant: Joueur, nom: str) -> Boudoir:
    """Crée un boudoir avec le gérant et le nom requis"""
    now = datetime.datetime.now()
    categ = await tools.multicateg(config.boudoirs_category_name)
    chan = await config.guild.create_text_channel(
        nom,
        topic=f"Boudoir crée le {now:%d/%m à %H:%M}. Gérant(e) : {gerant.nom}",
        category=categ,
    )

    boudoir = Boudoir(chan_id=chan.id, nom=nom, ts_created=now)
    boudoir.add()
    await add_joueur_to_boudoir(boudoir, gerant, gerant=True)
    await tools.log(f"Boudoir {chan.mention} créé par {gerant.nom}.")

    return boudoir


async def _invite(joueur: Joueur, boudoir: Boudoir, invite_msg: discord.Message) -> None:
    """Invitation d'un joueur dans un boudoir (lancer comme tâche à part)"""

    async def ack_invitation_response(info):
        try:
            await invite_msg.reply(info)
        except discord.HTTPException:  # Message d'invitation supprimé
            await boudoir.chan.send(info)

    class _InviteView(discord.ui.View):
        @discord.ui.button(style=discord.ButtonStyle.success, emoji="✅")
        async def ok(self, vote_interaction: discord.Interaction, button: discord.ui.Button):
            async with DiscordJourney(vote_interaction) as journey:
                if await add_joueur_to_boudoir(boudoir, joueur):
                    await ack_invitation_response(f"{joueur.nom} a rejoint le boudoir !")
                    await journey.send(f"Tu as bien rejoint {boudoir.chan.mention} !")
                else:
                    await journey.send("Impossible de rejoindre le boudoir :(")

        @discord.ui.button(style=discord.ButtonStyle.secondary, emoji="❌")
        async def c_non(self, contre_haro_interaction: discord.Interaction, button: discord.ui.Button):
            async with DiscordJourney(contre_haro_interaction) as journey:
                await ack_invitation_response(f"{joueur.nom} a refusé l'invitation à rejoindre ce boudoir.")
                await journey.send("Invitation refusée.")

    await joueur.private_chan.send(
        f"{joueur.member.mention} {boudoir.gerant.nom} t'as invité(e) à "
        f"rejoindre son boudoir : « {boudoir.nom} » !\nAcceptes-tu ?",
        view=_InviteView(),
    )


DESCRIPTION = """Gestion des salons"""

boudoir = app_commands.Group(name="boudoir", description="Gestion des boudoirs")
"""Gestion des boudoirs

Les options relatives à un boudoir précis ne peuvent être
exécutées que dans ce boudoir ; certaines sont réservées au
gérant dudit boudoir.
"""


@boudoir.command()
@tools.joueurs_only
@tools.private()
@journey_command
async def list(journey: DiscordJourney):
    """Liste les boudoirs dans lesquels tu es."""
    joueur = Joueur.from_member(journey.member)
    bouderies = joueur.bouderies

    if not bouderies:
        await journey.send("Tu n'es dans aucun boudoir pour le moment.\nUtilise `/boudoir create` pour en créer un.")
        return

    rep = "Tu es dans les boudoirs suivants :"
    for bouderie in bouderies:
        rep += f"\n - {bouderie.boudoir.chan.mention}"
        if bouderie.gerant:
            rep += " (gérant)"

    rep += "\n\nUtilise `/boudoir leave` dans un boudoir pour le quitter."

    await journey.send(rep)


@boudoir.command()
@tools.vivants_only
@tools.private()
@journey_command
async def create(journey: DiscordJourney, *, nom: app_commands.Range[str, 1, 32]):
    """Crée un nouveau boudoir dont tu es gérant(e).

    Args:
        nom: Le nom du boudoir à créer (doit faire moins de 32 caractères).
    """
    member = journey.member
    joueur = Joueur.from_member(member)

    boudoir = await _create_boudoir(joueur, nom)
    await boudoir.chan.send(
        f"{member.mention}, voici ton boudoir ! "
        "Tu peux maintenant y inviter des gens avec la commande `/boudoir invite`."
    )

    await journey.send(f"Ton boudoir a bien été créé : {boudoir.chan.mention}")


@boudoir.command()
@tools.joueurs_only
@journey_command
@in_boudoir
@gerant_only
async def invite(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.VivantTransformer]):
    """Invite un joueur à rejoindre ce boudoir (utilisable dans un boudoir que tu gères uniquement).

    Args:
        joueur: Le joueur à inviter dans ce boudoir.
    """
    boudoir = Boudoir.from_channel(journey.channel)
    if joueur in boudoir.joueurs:
        await journey.send(f":x: {joueur.nom} est déjà dans ce boudoir !")
        return

    mess = await journey.send(f"Invitation envoyée à {joueur.nom}.")
    asyncio.create_task(_invite(joueur, boudoir, mess))
    # On envoie l'invitation en arrière-plan (libération du chan).


@boudoir.command()
@tools.joueurs_only
@journey_command
@in_boudoir
@gerant_only
async def expulse(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.VivantTransformer]):
    """Expulse un membre de ce boudoir (utilisable dans un boudoir que tu gères uniquement).

    Args:
        joueur: Le joueur à expulser de ce boudoir.
    """
    boudoir = Boudoir.from_channel(journey.channel)
    if joueur not in boudoir.joueurs:
        await journey.send(f":x: {joueur.nom} n'est pas membre du boudoir !")
        return

    await remove_joueur_from_boudoir(boudoir, joueur)
    await joueur.private_chan.send(f"Tu as été expulsé(e) du boudoir « {boudoir.nom} ».")
    await journey.send(f"{joueur.nom} a bien été expulsé de ce boudoir.")


@boudoir.command()
@tools.joueurs_only
@journey_command
@in_boudoir
async def leave(journey: DiscordJourney):
    """Quitte ce boudoir (utilisable dans un boudoir uniquement)."""
    joueur = Joueur.from_member(journey.member)
    boudoir = Boudoir.from_channel(journey.channel)

    if boudoir.gerant == joueur:
        await journey.send(
            "Tu ne peux pas quitter un boudoir que tu gères.\n"
            "Utilise `/boudoir transfer` pour passer les droits de gestion ou `/boudoir delete` pour le supprimer."
        )
        return

    await journey.ok_cancel("Veux-tu vraiment quitter ce boudoir ? Tu ne pourras pas y retourner sans invitation.")

    await remove_joueur_from_boudoir(boudoir, joueur)
    await journey.send(tools.ital(f"{joueur.nom} a quitté ce boudoir."))


@boudoir.command()
@tools.joueurs_only
@journey_command
@in_boudoir
@gerant_only
async def transfer(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.VivantTransformer]):
    """Transfère les droits de gestion de ce boudoir (utilisable dans un boudoir que tu gères uniquement).

    Args:
        joueur: Le joueur à qui transférer ce boudoir. Doit en être membre.
    """
    boudoir = Boudoir.from_channel(journey.channel)
    if joueur not in boudoir.joueurs:
        await journey.send(f"{joueur.nom} n'est pas membre de ce boudoir !")
        return

    if not await journey.yes_no(
        "Veux-tu vraiment transférer les droits de ce boudoir ? Tu ne pourras pas les récupérer par toi-même."
    ):
        await journey.send("Mission aborted.")
        return

    boudoir.gerant = joueur
    boudoir.update()
    await boudoir.chan.edit(topic=f"Boudoir crée le {boudoir.ts_created:%d/%m à %H:%M}. Gérant(e) : {joueur.nom}")

    await journey.send(f"Boudoir transféré à {joueur.nom}.")


@boudoir.command()
@tools.joueurs_only
@journey_command
@in_boudoir
@gerant_only
async def delete(journey: DiscordJourney):
    """Supprime ce boudoir (utilisable dans un boudoir que tu gères uniquement)."""
    boudoir = Boudoir.from_channel(journey.channel)
    await journey.ok_cancel("Veux-tu vraiment supprimer ce boudoir ? Cette action est irréversible.")

    await journey.send("Suppression...")
    for joueur in boudoir.joueurs:
        await remove_joueur_from_boudoir(boudoir, joueur)
        await joueur.private_chan.send(f"Le boudoir « {boudoir.nom } » a été supprimé.")

    await boudoir.chan.edit(name=f"\N{CROSS MARK} {boudoir.nom}")
    await journey.send(
        tools.ital("[Tous les joueurs ont été exclus de ce boudoir ; le channel reste présent pour archive.]")
    )


@boudoir.command()
@tools.joueurs_only
@journey_command
@in_boudoir
@gerant_only
async def rename(journey: DiscordJourney, *, nom: app_commands.Range[str, 1, 32]):
    """Renomme ce boudoir (utilisable dans un boudoir que tu gères uniquement).

    Args:
        nom: Le nouveau nom de ce boudoir (doit faire moins de 32 caractères).
    """
    boudoir = Boudoir.from_channel(journey.channel)

    boudoir.nom = nom
    boudoir.update()
    await boudoir.chan.edit(name=nom)
    await journey.send("Boudoir renommé avec succès.")


@boudoir.command()
@tools.joueurs_only
@journey_command
@in_boudoir
@gerant_only
async def ping(journey: DiscordJourney, *, message: str = ""):
    """Mentionne tous les joueurs vivants dans le boudoir (utilisable dans un boudoir que tu gères uniquement).

    Args:
        message: Le message à faire passer, optionnel
    """
    await journey.send(f"{config.Role.joueur_en_vie.mention} {message}")


@boudoir.command()
@tools.mjs_only
@journey_command
async def find(
    journey: DiscordJourney,
    joueur1: app_commands.Transform[Joueur, tools.VivantTransformer],
    joueur2: app_commands.Transform[Joueur, tools.VivantTransformer] | None = None,
    joueur3: app_commands.Transform[Joueur, tools.VivantTransformer] | None = None,
):
    """Trouve le(s) boudoir(s) réunissant certains joueurs (COMMANDE MJ)

    Args:
        joueur1: Premier joueur
        joueur1: Deuxième joueur (optionnel)
        joueur1: Troisième joueur (optionnel)
    """
    joueurs = [joueur1]
    if joueur2:
        joueurs.append(joueur2)
    if joueur3:
        joueurs.append(joueur3)
    boudoirs = [boudoir for boudoir in Boudoir.query.all() if all(joueur in boudoir.joueurs for joueur in joueurs)]

    if not boudoirs:
        await journey.send(f":x: Pas de boudoir(s) réunissant {joueurs}.")
    else:
        liste = "\n".join(f"- {boudoir.chan.mention} ({len(boudoir.joueurs)} joueurs)" for boudoir in boudoirs)
        await journey.send(f"{len(boudoirs)} boudoirs :\n{liste}")


async def _mp(journey: DiscordJourney, *, joueur: Joueur):
    member = journey.member
    moi = Joueur.from_member(member)

    if moi == joueur:
        await journey.send(f"Ton boudoir a bien été créé : {moi.private_chan.mention}\n(tocard)")
        return

    # Recherche si boudoir existant
    boudoir = next((boudoir for boudoir in moi.boudoirs if set(boudoir.joueurs) == {moi, joueur}), None)
    if boudoir:
        await journey.send(f":x: Ce boudoir existe déjà : {boudoir.chan.mention}")
        return

    boudoir = await _create_boudoir(moi, f"{moi.nom} × {joueur.nom}")
    await boudoir.chan.send(f"{member.mention}, voici ton boudoir !")

    await journey.send(f":sunglasses: Ton boudoir a bien été créé : {boudoir.chan.mention}")

    mess = await boudoir.chan.send(f"Invitation envoyée à {joueur.nom}.")
    asyncio.create_task(_invite(joueur, boudoir, mess))
    # On envoie l'invitation en arrière-plan (libération du chan).


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def mp(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.VivantTransformer]):
    """Raccourci pour créer un boudoir et y ajouter un joueur.

    Si un boudoir existe déjà avec uniquement ce joueur et toi, n'en crée pas un nouveau.

    Args:
        joueur: La personne avec qui créer un boudoir.
    """
    await _mp(journey, joueur=joueur)


@app_commands.context_menu(name="Créer un boudoir avec ce joueur")
@tools.vivants_only
@journey_context_menu
async def mp_menu(journey: DiscordJourney, member: discord.Member):
    if member.top_role >= config.Role.mj:
        await journey.send(":x: Tu ne peux pas créer de boudoir avec un MJ. Ta conversation privée est là pour ça !")
        return

    if member == config.bot.user:
        await journey.send(":x: Tu ne peux pas créer de boudoir avec le bot, sacrebleu !")
        return

    try:
        joueur = Joueur.from_member(member)
    except ValueError:
        await journey.send(":x: Hmm, ce joueur n'a pas l'air inscrit, réessaye plus tard !")
        return

    await _mp(journey, joueur=joueur)


@app_commands.command()
@tools.mjs_only
@journey_command
async def addhere(
    journey: DiscordJourney,
    crit: Literal["chambre", "statut", "camp", "role", "votant_village", "votant_loups", "role_actif"],
    filtre: str,
):
    """Ajoute les membres au chan courant (COMMANDE MJ)

    Args:
        crit: Critère sur lequel filtrer les joueurs à ajouter.
        filtre: Valeur de `crit` à récupérer.
    """
    ts_debut = journey.created_at - datetime.timedelta(microseconds=1)

    col = Joueur.attrs[crit]
    arg = transtype(filtre.strip(), col)
    joueurs = Joueur.query.filter_by(**{crit: arg}).all()

    for joueur in joueurs:
        await journey.channel.set_permissions(joueur.member, read_messages=True)
        await journey.channel.send(f"{joueur.nom} ajouté")

    await journey.ok_cancel("Fini, purge les messages ?")
    await journey.channel.purge(after=ts_debut)


@app_commands.command()
@tools.mjs_only
@journey_command
async def purge(journey: DiscordJourney, limit: int | None = None):
    """Supprime tous les messages de ce chan (COMMANDE MJ)

    Args:
        limit: Nombre de messages à supprimer (défaut : tous).
    """
    if limit:
        mess = f"Supprimer les {limit} messages les plus récents de ce chan ?"
    else:
        mess = "Supprimer tous les messages de ce chan ?"

    journey.last_interaction()
    await journey.ok_cancel(mess)
    await journey.channel.purge(limit=limit)
